#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *


def test_type(name, signal):
    return signal == fis_types[name]

# Transport TX -------------------------------------------------------------------------------------

class LiteSATATransportTX(Module):
    def __init__(self, link):
        self.sink = sink = stream.Endpoint(transport_tx_description(32))

        # # #

        cmd_ndwords = max(fis_reg_h2d_header.length, fis_data_header.length)
        encoded_cmd = Signal(cmd_ndwords*32)

        counter       = Signal(max=cmd_ndwords+1)
        counter_ce    = Signal()
        counter_reset = Signal()
        self.sync += \
            If(counter_reset,
                counter.eq(0)
            ).Elif(counter_ce,
                counter.eq(counter + 1)
            )

        cmd_len       = Signal(len(counter))
        cmd_with_data = Signal()

        cmd_send  = Signal()
        data_send = Signal()
        cmd_done  = Signal()

        fis_type        = Signal(8)
        update_fis_type = Signal()

        def test_type_tx(name):
            return test_type(name, sink.type)

        self.fsm = fsm = FSM(reset_state="IDLE")
        self.submodules += fsm
        fsm.act("IDLE",
            sink.ready.eq(0),
            counter_reset.eq(1),
            update_fis_type.eq(1),
            If(sink.valid,
                If(test_type_tx("REG_H2D"),
                    NextState("SEND_CTRL_CMD")
                ).Elif(test_type_tx("DATA"),
                    NextState("SEND_DATA_CMD")
                ).Else(
                    sink.ready.eq(1)
                )
            ).Else(
                sink.ready.eq(1)
            )
        )
        self.sync += \
            If(update_fis_type, fis_type.eq(link.source.data[:8]))

        fsm.act("SEND_CTRL_CMD",
            fis_reg_h2d_header.encode(sink, encoded_cmd),
            cmd_len.eq(fis_reg_h2d_header.length-1),
            cmd_send.eq(1),
            If(cmd_done,
                sink.ready.eq(1),
                NextState("IDLE")
            )
        )
        fsm.act("SEND_DATA_CMD",
            sink.ready.eq(0),
            fis_data_header.encode(sink, encoded_cmd),
            cmd_len.eq(fis_data_header.length-1),
            cmd_with_data.eq(1),
            cmd_send.eq(1),
            If(cmd_done,
                NextState("SEND_DATA")
            )
        )
        fsm.act("SEND_DATA",
            data_send.eq(1),
            sink.ready.eq(link.sink.ready),
            If(sink.valid & sink.last & sink.ready,
                NextState("IDLE")
            )
        )

        cmd_cases = {}
        for i in range(cmd_ndwords):
            cmd_cases[i] = [link.sink.data.eq(encoded_cmd[32*i:32*(i+1)])]

        self.comb += [
            counter_ce.eq(sink.valid & link.sink.ready),
            cmd_done.eq((counter == cmd_len) &
                        link.sink.valid &
                        link.sink.ready),
            If(cmd_send,
                link.sink.valid.eq(sink.valid),
                link.sink.last.eq((counter == cmd_len) & ~cmd_with_data),
                Case(counter, cmd_cases)
            ).Elif(data_send,
                link.sink.valid.eq(sink.valid),
                link.sink.last.eq(sink.last),
                link.sink.data.eq(sink.data)
            )
        ]

# Transport RX -------------------------------------------------------------------------------------

class LiteSATATransportRX(Module):
    def __init__(self, link):
        self.source = source = stream.Endpoint(transport_rx_description(32))

        # # #

        cmd_ndwords = max(fis_reg_d2h_header.length,
                          fis_dma_activate_d2h_header.length,
                          fis_pio_setup_d2h_header.length,
                          fis_data_header.length)
        encoded_cmd = Signal(cmd_ndwords*32)

        counter       = Signal(max=cmd_ndwords+1)
        counter_ce    = Signal()
        counter_reset = Signal()
        self.sync += \
            If(counter_reset,
                counter.eq(0)
            ).Elif(counter_ce,
                counter.eq(counter + 1)
            )

        cmd_len = Signal(len(counter))

        cmd_receive  = Signal()
        data_receive = Signal()
        cmd_done     = Signal()
        data_done    = Signal()

        def test_type_rx(name):
            return test_type(name, link.source.data[:8])

        self.fsm = fsm = FSM(reset_state="IDLE")
        self.submodules += fsm

        fis_type        = Signal(8)
        update_fis_type = Signal()

        fsm.act("IDLE",
            link.source.ready.eq(0),
            counter_reset.eq(1),
            update_fis_type.eq(1),
            If(link.source.valid,
                If(test_type_rx("REG_D2H"),
                    NextState("RECEIVE_CTRL_CMD")
                ).Elif(test_type_rx("DMA_ACTIVATE_D2H"),
                    NextState("RECEIVE_CTRL_CMD")
                ).Elif(test_type_rx("PIO_SETUP_D2H"),
                    NextState("RECEIVE_CTRL_CMD")
                ).Elif(test_type_rx("DATA"),
                    NextState("RECEIVE_DATA_CMD"),
                ).Else(
                    link.source.ready.eq(1)
                )
            ).Else(
                link.source.ready.eq(1)
            )
        )
        self.sync += \
            If(update_fis_type, fis_type.eq(link.source.data[:8]))

        fsm.act("RECEIVE_CTRL_CMD",
            If(test_type("REG_D2H", fis_type),
                cmd_len.eq(fis_reg_d2h_header.length-1)
            ).Elif(test_type("DMA_ACTIVATE_D2H", fis_type),
                cmd_len.eq(fis_dma_activate_d2h_header.length-1)
            ).Else(
                cmd_len.eq(fis_pio_setup_d2h_header.length-1)
            ),
            cmd_receive.eq(1),
            link.source.ready.eq(1),
            If(cmd_done,
                NextState("PRESENT_CTRL_CMD")
            )
        )
        fsm.act("PRESENT_CTRL_CMD",
            source.valid.eq(1),
            source.last.eq(1),
            If(test_type("REG_D2H", fis_type),
                fis_reg_d2h_header.decode(encoded_cmd, source)
            ).Elif(test_type("DMA_ACTIVATE_D2H", fis_type),
                fis_dma_activate_d2h_header.decode(encoded_cmd, source)
            ).Else(
                fis_pio_setup_d2h_header.decode(encoded_cmd, source)
            ),
            If(source.valid & source.ready,
                NextState("IDLE")
            )
        )
        fsm.act("RECEIVE_DATA_CMD",
            cmd_len.eq(fis_data_header.length-1),
            cmd_receive.eq(1),
            link.source.ready.eq(1),
            If(cmd_done,
                NextState("PRESENT_DATA")
            )
        )
        fsm.act("PRESENT_DATA",
            data_receive.eq(1),
            source.valid.eq(link.source.valid),
            fis_data_header.decode(encoded_cmd, source),
            source.last.eq(link.source.last),
            source.error.eq(link.source.error),
            source.data.eq(link.source.data),
            link.source.ready.eq(source.ready),
            If(source.valid & source.last & source.ready,
                NextState("IDLE")
            )
        )

        cmd_cases = {}
        for i in range(cmd_ndwords):
            cmd_cases[i] = [encoded_cmd[32*i:32*(i+1)].eq(link.source.data)]

        self.comb += \
            If(cmd_receive & link.source.valid,
                counter_ce.eq(1)
            )
        self.sync += \
            If(cmd_receive,
                Case(counter, cmd_cases),
            )
        self.comb += cmd_done.eq((counter == cmd_len) & link.source.ready)

# Transport ----------------------------------------------------------------------------------------

class LiteSATATransport(Module):
    def __init__(self, link):
        self.submodules.tx = LiteSATATransportTX(link)
        self.submodules.rx = LiteSATATransportRX(link)
        self.sink, self.source = self.tx.sink, self.rx.source



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from functools import reduce
from operator import and_, or_

from litesata.common import *
from litesata.frontend.arbitration import LiteSATAUserPort

from litex.soc.interconnect.packet import Status, Arbiter, Dispatcher

# LiteSATA Striping --------------------------------------------------------------------------------

class LiteSATAStripingTX(Module):
    """SATA Striping TX

    Split cmds and writes data on N different controllers.

    This module provides a mirroring_mode that is used by the mirroring module to
    dispatch identicals writes to the controllers. This avoid code duplication in
    between striping and mirroring modules. In this special case, port's data width
    is dw (same as controllers)
    """
    def __init__(self, n, dw, mirroring_mode=False):
        self.sink = sink = stream.Endpoint(command_tx_description(dw*n if not mirroring_mode else dw))
        self.sources = sources = [stream.Endpoint(command_tx_description(dw)) for i in range(n)]

        # # #

        split = Signal()

        already_acked = Signal(n)
        self.sync += If(split & sink.valid,
                already_acked.eq(already_acked | Cat(*[s.ready for s in sources])),
                If(sink.ready, already_acked.eq(0))
            )

        self.fsm = fsm = FSM(reset_state="IDLE")
        self.submodules += fsm
        fsm.act("IDLE",
            sink.ready.eq(0),
            If(sink.valid,
                NextState("SPLIT")
            ).Else(
                sink.ready.eq(1)
            )
        )

        # split data and ctrl signals (except valid & ready managed in fsm)
        for i, s in enumerate(sources):
            self.comb += sink.connect(s, omit=set(["valid", "ready", "data"]))
            if mirroring_mode:
                self.comb += s.data.eq(sink.data)
            else:
                self.comb += s.data.eq(sink.data[i*dw:(i+1)*dw])

        fsm.act("SPLIT",
            split.eq(1),
            [s.valid.eq(sink.valid & ~already_acked[i]) for i, s in enumerate(sources)],
            sink.ready.eq(reduce(and_, [s.ready | already_acked[i] for i, s in enumerate(sources)])),
            If(sink.valid & sink.last & sink.ready,
                NextState("IDLE")
            )
        )

class LiteSATAStripingRX(Module):
    """SATA Striping RX

    Combine acknowledges and reads data from N different controllers.

    This module provides a mirroring_mode that is used by the mirroring module to
    dispatch identicals writes to the controllers. This avoid code duplication in
    between striping and mirroring modules. In this special case, port's data width
    is dw (same as controllers)
    """
    def __init__(self, n, dw, mirroring_mode=False):
        self.sinks = sinks = [stream.Endpoint(command_rx_description(dw)) for i in range(n)]
        self.source = source = stream.Endpoint(command_rx_description(dw*n if not mirroring_mode else dw))

        # # #

        valid_all = Signal()
        self.comb += valid_all.eq(reduce(and_, [s.valid for s in sinks]))

        self.fsm = fsm = FSM(reset_state="IDLE")
        self.submodules += fsm
        fsm.act("IDLE",
            If(valid_all,
                NextState("COMBINE")
            )
        )

        # use first sink for ctrl signals (except for valid, ready & failed)
        self.comb += sinks[0].connect(source, omit=set(["valid", "ready", "failed", "data"]))
        # combine datas
        if mirroring_mode:
            self.comb += source.data.eq(0) # mirroring only used for writes
        else:
            for i, s in enumerate(sinks):
                self.comb += source.data[i*dw:(i+1)*dw].eq(s.data)


        fsm.act("COMBINE",
            source.failed.eq(reduce(or_, [s.failed for s in sinks])), # XXX verify this is enough
            source.valid.eq(reduce(and_, [s.valid for s in sinks])),
            [s.ready.eq(source.valid & source.ready) for s in sinks],
            If(source.valid & source.last & source.ready,
                NextState("IDLE")
            )
        )


class LiteSATAStriping(Module):
    """SATA Striping

    Segment data so that data is stored on N different controllers.
                     +----> controller0 (dw)
    port (N*dw) <----+----> controllerX (dw)
                     +----> controllerN (dw)

    Characteristics:
        - port's visible capacity = N x controller's visible capacity
        - port's throughput = N x (slowest) controller's throughput

    Can be used to increase capacity and writes/reads throughput.
    """
    def __init__(self, controllers):

        # # #

        n  = len(controllers)
        dw = len(controllers[0].sink.data)

        self.submodules.tx = LiteSATAStripingTX(n, dw)
        self.submodules.rx = LiteSATAStripingRX(n, dw)
        for i in range(n):
            self.comb += [
                self.tx.sources[i].connect(controllers[i].sink),
                controllers[i].source.connect(self.rx.sinks[i])
            ]
        self.sink, self.source = self.tx.sink, self.rx.source

# LiteSATA Mirroring -------------------------------------------------------------------------------

class LiteSATAMirroringCtrl(Module):
    def __init__(self, n):
        self.new_cmds    = Signal(n)
        self.ack_cmds    = Signal(n)

        self.reading     = Signal()
        self.writing     = Signal()

        self.wants_write = Signal()
        self.write_sel   = Signal(max=n)

        # # #

        pending_cmds = Signal(n)
        self.sync += pending_cmds.eq(self.new_cmds | (pending_cmds & ~self.ack_cmds))
        can_commute = Signal()
        self.comb += can_commute.eq((pending_cmds | self.new_cmds)  == 0)

        self.fsm = fsm = FSM(reset_state="READ")
        self.submodules += fsm
        fsm.act("READ",
            self.reading.eq(1),
            If(self.wants_write & can_commute,
                NextState("WRITE")
            )
        )
        fsm.act("WRITE",
            self.writing.eq(1),
            If(~self.wants_write & can_commute,
                NextState("READ")
            )
        )


class LiteSATAMirroringTX(Module):
    def __init__(self, n, dw, ctrl):
        self.sinks = sinks = [stream.Endpoint(command_tx_description(dw)) for i in range(n)]
        self.sources = sources = [stream.Endpoint(command_tx_description(dw)) for i in range(n)]

        # # #

        wants_write = Signal()

        reading = Signal()
        writing = Signal()

        reads  = [stream.Endpoint(command_tx_description(dw)) for i in range(dw)]
        writes = [stream.Endpoint(command_tx_description(dw)) for i in range(dw)]
        for sink, read, write in zip(sinks, reads, writes):
            read_stall  = Signal()
            read_status = Status(read)
            self.submodules += read_status
            self.comb += [
                sink.connect(read, omit=set(["valid", "ready"])),
                sink.connect(write, omit=set(["valid", "ready"])),
                read.valid.eq(sink.valid & (sink.read | sink.identify) & ~read_stall),
                write.valid.eq(sink.valid & sink.write),
                If(sink.read | sink.identify,
                    sink.ready.eq((read.ready & ~read_stall))
                ).Else(
                    sink.ready.eq(write.ready)
                )
            ]
            self.sync += \
                If(~ctrl.wants_write,
                    read_stall.eq(0)
                ).Elif(~read_status.ongoing,
                    read_stall.eq(1)
                )

        write_striper = LiteSATAStripingTX(n, dw, mirroring_mode=True)
        write_arbiter = Arbiter(writes, write_striper.sink)
        self.submodules += write_striper, write_arbiter

        for i in range(n):
            source_status = Status(sources[i])
            self.submodules += source_status
            self.comb += [
                If(ctrl.reading,
                    reads[i].connect(sources[i]) # independent reads
                ).Elif(ctrl.writing,
                    write_striper.sources[i].connect(sources[i]) # identical writes
                ),
                ctrl.new_cmds[i].eq(source_status.last)
            ]
        write_striper_sink_status = Status(write_striper.sink)
        self.submodules += write_striper_sink_status
        self.comb += [
            ctrl.wants_write.eq(write_striper_sink_status.ongoing),
            ctrl.write_sel.eq(write_arbiter.rr.grant)
        ]


class LiteSATAMirroringRX(Module):
    def __init__(self, n, dw, ctrl):
        self.sinks = sinks = [stream.Endpoint(command_rx_description(dw)) for i in range(n)]
        self.sources = sources = [stream.Endpoint(command_rx_description(dw)) for i in range(n)]

        # # #

        muxs = [Multiplexer(command_rx_description(dw), 2) for i in range(n)]
        self.submodules += muxs

        writes = [mux.sink0 for mux in muxs]
        reads  = [mux.sink1 for mux in muxs]

        for mux, source in zip(muxs, sources):
            self.comb += [
                mux.sel.eq(ctrl.reading),
                mux.source.connect(source)
            ]

        write_striper = LiteSATAStripingRX(n, dw, mirroring_mode=True)
        write_dispatcher = Dispatcher(write_striper.source, writes)
        self.comb += write_dispatcher.sel.eq(ctrl.write_sel)
        self.submodules += write_striper, write_dispatcher

        for i in range(n):
            sink_status = Status(sinks[i])
            self.submodules += sink_status
            self.comb += [
                sinks[i].connect(reads[i], omit=set(["valid", "ready"])),
                sinks[i].connect(write_striper.sinks[i], omit=set(["valid", "ready"])),
                reads[i].valid.eq(sinks[i].valid & ctrl.reading),
                write_striper.sinks[i].valid.eq(sinks[i].valid & ctrl.writing),
                sinks[i].ready.eq(reads[i].ready | write_striper.sinks[i].ready),
                ctrl.ack_cmds[i].eq(sink_status.last & sinks[i].end)
            ]


class LiteSATAMirroring(Module):
    """SATA Mirroring

    The mirroring module handles N controllers and provides N ports.
    Each port has its dedicated controller for reads:
        port0 <----> controller0
        portX <----> controllerX
        portN <----> controllerN

    Writes are mirrored on each controller:
                   (port0 write)           |            (portN write)
        port0 ----------+----> controller0 | port0 (stalled) +-----> controller0
        portX (stalled) +----> controllerX | portX (stalled) +-----> controllerX
        portN (stalled) +----> controllerN | portN ----------+-----> controllerN

    Writes have priority on reads. When a write is presented on one of the port, the
    module waits for all ongoing reads to finish and commute to write mode. Once all writes are
    serviced it returns to read mode.

    Characteristics:
        - port's visible capacity = controller's visible capacity
        - total writes throughput = (slowest) controller's throughput
        - total reads throughput = N x controller's throughput

    Can be used for data redundancy and/or to increase total reads speed.
    """
    def __init__(self, controllers):
        n  = len(controllers)
        dw = len(controllers[0].sink.data)
        self.ports = [LiteSATAUserPort(dw) for i in range(n)]

        # # #

        self.submodules.ctrl = LiteSATAMirroringCtrl(n)
        self.submodules.tx   = LiteSATAMirroringTX(n, dw, self.ctrl)
        self.submodules.rx   = LiteSATAMirroringRX(n, dw, self.ctrl)
        for i in range(n):
            self.comb += [
                self.ports[i].sink.connect(self.tx.sinks[i]),
                self.tx.sources[i].connect(controllers[i].sink),

                controllers[i].source.connect(self.rx.sinks[i]),
                self.rx.sources[i].connect(self.ports[i].source)
            ]



#
# This file is part of LiteSATA.
#
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2023 Gabriel L. Somlo <gsomlo@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

from math import log2

from migen import *

from litex.gen import *
from litex.soc.interconnect.csr import *
from litex.soc.interconnect import stream

from litex.soc.cores.dma import WishboneDMAWriter, WishboneDMAReader

from litesata.common import logical_sector_size

# SATA Sector2Mem DMA ------------------------------------------------------------------------------

class LiteSATASector2MemDMA(Module, AutoCSR):
    """Sector to Memory DMA

    Read a sector from the SATA core and write it to memory through DMA.
    """
    def __init__(self, port, bus, endianness="little"):
        self.port     = port
        self.bus      = bus
        self.sector   = CSRStorage(48)
        self.nsectors = CSRStorage(16)
        self.base     = CSRStorage(64)
        self.start    = CSR()
        self.done     = CSRStatus()
        self.error    = CSRStatus()
        self.irq      = Signal()

        # # #

        port_bytes = port.dw//8
        dma_bytes  = bus.data_width//8
        count      = Signal(max=logical_sector_size//dma_bytes)
        crt_sec    = Signal(48)
        crt_base   = Signal(64)

        # Sector buffer
        buf = stream.SyncFIFO([("data", port.dw)], logical_sector_size//port_bytes)
        self.submodules.buf = buf

        # Converter
        conv = stream.Converter(nbits_from=port.dw, nbits_to=bus.data_width)
        self.submodules.conv = conv

        # Connect Port to Sector Buffer
        self.comb += port.source.connect(buf.sink, keep={"valid", "ready", "last", "data"})

        # Connect Sector Buffer to Converter
        self.comb += buf.source.connect(conv.sink)

        # DMA
        dma = WishboneDMAWriter(bus, with_csr=False, endianness=endianness)
        self.submodules.dma = dma

        # Control FSM
        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            If(self.start.re,
                NextValue(count,             0),
                NextValue(crt_sec,           self.sector.storage),
                NextValue(crt_base,          self.base.storage),
                NextValue(self.error.status, 0),
                NextState("SEND-CMD")
            ).Else(
                self.done.status.eq(1)
            ),
            conv.source.ready.eq(1)
        )
        fsm.act("SEND-CMD",
            # Send read command for 1 Sector.
            port.sink.valid.eq(1),
            port.sink.last.eq(1),
            port.sink.read.eq(1),
            port.sink.sector.eq(crt_sec),
            port.sink.count.eq(1),
            If(port.sink.ready,
                NextState("RECEIVE-DATA-DMA")
            )
        )
        fsm.act("RECEIVE-DATA-DMA",
            # Connect Converter to DMA.
            dma.sink.valid.eq(conv.source.valid),
            dma.sink.last.eq(conv.source.last),
            dma.sink.address.eq(crt_base[int(log2(dma_bytes)):] + count),
            dma.sink.data.eq(reverse_bytes(conv.source.data)),
            conv.source.ready.eq(dma.sink.ready),
            If(dma.sink.valid & dma.sink.ready,
                NextValue(count, count + 1),
                If(dma.sink.last,
                    self.irq.eq(1),
                    NextState("SECTOR-LOOP")
                )
            ),

            # Monitor errors
            If(port.source.valid & port.source.ready & port.source.failed,
                self.irq.eq(1),
                NextValue(self.error.status, 1),
                NextState("IDLE"),
            )
        )
        fsm.act("SECTOR-LOOP",
            If(crt_sec == (self.sector.storage + self.nsectors.storage - 1),
                self.irq.eq(1),
                NextState("IDLE")
            ).Else(
                NextValue(count,    0),
                NextValue(crt_sec,  crt_sec + 1),
                NextValue(crt_base, crt_base + 512),
                NextValue(self.error.status, 0),
                conv.source.ready.eq(1),
                NextState("SEND-CMD")
            )
        )

# SATA Mem2Sector DMA ------------------------------------------------------------------------------

class LiteSATAMem2SectorDMA(Module, AutoCSR):
    """Memory 2 Sector DMA

    Read memory through DMA and write it to a sector of the SATA core.
    """
    def __init__(self, bus, port, endianness="little"):
        self.bus      = bus
        self.port     = port
        self.sector   = CSRStorage(48)
        self.nsectors = CSRStorage(16)
        self.base     = CSRStorage(64)
        self.start    = CSR()
        self.done     = CSRStatus()
        self.error    = CSRStatus()
        self.irq      = Signal()

        # # #

        dma_bytes  = bus.data_width//8
        port_bytes = port.dw//8
        count      = Signal(max=logical_sector_size//min(dma_bytes, port_bytes))
        crt_sec    = Signal(48)
        crt_base   = Signal(64)

        # DMA
        dma = WishboneDMAReader(bus, with_csr=False, endianness=endianness)
        self.submodules.dma = dma

        # Sector buffer
        buf = stream.SyncFIFO([("data", port.dw)], logical_sector_size//dma_bytes)
        self.submodules.buf = buf

        # Converter
        conv = stream.Converter(nbits_from=bus.data_width, nbits_to=port.dw)
        self.submodules.conv = conv

        # Connect DMA to Sector Buffer
        self.comb += dma.source.connect(buf.sink)

        # Connect Sector Buffer to Converter
        self.comb += buf.source.connect(conv.sink)

        # Control FSM
        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            If(self.start.re,
                NextValue(count,             0),
                NextValue(crt_sec,           self.sector.storage),
                NextValue(crt_base,          self.base.storage),
                NextValue(self.error.status, 0),
                NextState("READ-DATA-DMA")
            ).Else(
                self.done.status.eq(1)
            ),
            conv.source.ready.eq(1)
        )
        fsm.act("READ-DATA-DMA",
            # Read Sector data over DMA.
            dma.sink.valid.eq(1),
            dma.sink.address.eq(crt_base[int(log2(dma_bytes)):] + count),
            If(dma.sink.valid & dma.sink.ready,
                NextValue(count, count + 1),
                If(count == (logical_sector_size//dma_bytes - 1),
                    NextValue(count, 0),
                    NextState("SEND-CMD-AND-DATA")
                )
            )
        )
        fsm.act("SEND-CMD-AND-DATA",
            # Send write command/data for 1 Sector.
            port.sink.valid.eq(1),
            port.sink.last.eq(count == (logical_sector_size//port_bytes - 1)),
            port.sink.write.eq(1),
            port.sink.sector.eq(crt_sec),
            port.sink.count.eq(1),
            port.sink.data.eq(reverse_bytes(conv.source.data)),
            If(port.sink.ready,
                conv.source.ready.eq(1),
                NextValue(count, count + 1),
                If(port.sink.last,
                    NextState("WAIT-ACK")
                )
            ),

            # Monitor errors
            port.source.ready.eq(1),
            If(port.source.valid & port.source.ready & port.source.failed,
                self.irq.eq(1),
                NextValue(self.error.status, 1),
                NextState("IDLE"),
            )
        )
        fsm.act("WAIT-ACK",
            port.source.ready.eq(1),
            If(port.source.valid,
                If(port.source.failed,
                    NextValue(self.error.status, 1),
                    self.irq.eq(1),
                    NextState("IDLE")
                ).Elif(crt_sec == (self.sector.storage + self.nsectors.storage - 1),
                    self.irq.eq(1),
                    NextState("IDLE")
                ).Else(
                    NextValue(count,    0),
                    NextValue(crt_sec,  crt_sec + 1),
                    NextValue(crt_base, crt_base + 512),
                    NextState("READ-DATA-DMA")
                )
            )
        )



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2016 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from litesata.common import *
from litesata.core.link import LiteSATACONTInserter, LiteSATACONTRemover

from litex.soc.interconnect.stream_sim import *


class ContPacket(list):
    def __init__(self, data=[]):
        self.ongoing = False
        self.done    = False
        for d in data:
            self.append(d)


class ContStreamer(PacketStreamer):
    def __init__(self):
        PacketStreamer.__init__(self, phy_description(32), packet_cls=ContPacket)

    @passive
    def generator(self):
        yield self.source.charisk.eq(0)
        # Note: for simplicity we generate charisk by detecting primitives in data
        for k, v in primitives.items():
            try:
                if self.source_data == v:
                    yield self.source.charisk.eq(0b0001)
            except:
                pass
        yield from PacketStreamer.generator(self)

class ContLogger(PacketLogger):
    def __init__(self):
        PacketLogger.__init__(self, phy_description(32), packet_cls=ContPacket)

class TestLinkCont(unittest.TestCase):
    def test_link_cont(self):
        def generator(dut):
            test_packet = ContPacket([
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["ALIGN"],
                primitives["ALIGN"],
                primitives["SYNC"],
                primitives["SYNC"],
                #primitives["SYNC"],
                0x00000000,
                0x00000001,
                0x00000002,
                0x00000003,
                0x00000004,
                0x00000005,
                0x00000006,
                0x00000007,
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["ALIGN"],
                primitives["ALIGN"],
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["SYNC"],
                primitives["SYNC"]]*4
                )
            streamer_packet = ContPacket(test_packet)
            yield from dut.streamer.send_blocking(streamer_packet)
            yield from dut.logger.receive(len(test_packet))
            #for d in self.logger.packet:
            #    print("{:08x}".format(d))

            # check results
            s, l, e = check(streamer_packet, dut.logger.packet)
            print("shift " + str(s) + " / length " + str(l) + " / errors " + str(e))
            self.assertEqual(s, 0)
            self.assertEqual(e, 0)

        class DUT(Module):
            def __init__(self):
                self.submodules.streamer            = ContStreamer()
                self.submodules.streamer_randomizer = Randomizer(phy_description(32), level=50)
                self.submodules.inserter            = LiteSATACONTInserter(phy_description(32))
                self.submodules.remover             = LiteSATACONTRemover(phy_description(32))
                self.submodules.logger_randomizer   = Randomizer(phy_description(32), level=50)
                self.submodules.logger              = ContLogger()

                self.submodules.pipeline = Pipeline(
                    self.streamer,
                    self.streamer_randomizer,
                    self.inserter,
                    self.remover,
                    self.logger_randomizer,
                    self.logger
                )

        dut = DUT()
        generators = {
            "sys" :   [generator(dut),
                       dut.streamer.generator(),
                       dut.streamer_randomizer.generator(),
                       dut.logger.generator(),
                       dut.logger_randomizer.generator()]
        }
        clocks = {"sys": 10}
        run_simulation(dut, generators, clocks)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020 Qui Nguyen <qui@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *
from litesata.common import _PulseSynchronizer, _RisingEdge

from migen.genlib.cdc import MultiReg
from migen.genlib.resetsync import AsyncResetSynchronizer

from liteiclink.serdes.common import *
from liteiclink.serdes.gty_ultrascale_init import GTYTXInit, GTYRXInit


class USPLiteSATAPHYCRG(Module):
    def __init__(self, refclk, pads, gty, gen):
        self.tx_reset = Signal()
        self.rx_reset = Signal()

        self.clock_domains.cd_sata_tx = ClockDomain()
        self.clock_domains.cd_sata_rx = ClockDomain()

        # CPLL -------------------------------------------------------------------------------------
        #   (gen3) 150MHz / VCO @ 3GHz / Linerate @ 6Gbps
        #   (gen2 & gen1) VCO still @ 3 GHz, Linerate is decreased with output dividers.
        if isinstance(refclk, (Signal, ClockSignal)):
            self.refclk = refclk
        else:
            self.refclk = Signal()
            self.specials += Instance("IBUFDS_GTE4",
                i_CEB = 0,
                i_I   = pads.clk_p,
                i_IB  = pads.clk_n,
                o_O   = self.refclk
            )
        self.comb += gty.refclk.eq(self.refclk)

        # TX clocking ------------------------------------------------------------------------------
        #   (gen3) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 300MHz (16-bits) /  150MHz (32-bits)
        #   (gen2) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 150MHz (16-bits) /   75MHz (32-bits)
        #   (gen1) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 75MHz  (16-bits) / 37.5MHz (32-bits)
        self.specials += Instance("BUFG_GT", i_I=gty.txoutclk, o_O=self.cd_sata_tx.clk)

        self.comb += gty.txusrclk.eq(self.cd_sata_tx.clk)
        self.comb += gty.txusrclk2.eq(self.cd_sata_tx.clk)

        # RX clocking ------------------------------------------------------------------------------
        #   (gen3) sata_rx recovered clk @ 300MHz (16-bits) /  150MHz (32-bits) from GTY RXOUTCLK
        #   (gen2) sata_rx recovered clk @ 150MHz (16-bits) /   75MHz (32-bits) from GTY RXOUTCLK
        #   (gen1) sata_rx recovered clk @ 75MHz  (16-bits) / 37.5MHz (32-bits) from GTY RXOUTCLK
        self.specials += Instance("BUFG_GT", i_I=gty.rxoutclk, o_O=self.cd_sata_rx.clk)

        self.comb += gty.rxusrclk.eq(self.cd_sata_rx.clk)
        self.comb += gty.rxusrclk2.eq(self.cd_sata_rx.clk)

        # Reset for SATA TX/RX clock domains -------------------------------------------------------
        self.specials += [
            AsyncResetSynchronizer(self.cd_sata_tx, ~gty.cplllock | self.tx_reset),
            AsyncResetSynchronizer(self.cd_sata_rx, ~gty.cplllock | self.rx_reset)
        ]

# --------------------------------------------------------------------------------------------------

class USPLiteSATAPHY(Module):
    def __init__(self, pads, gen, clk_freq, data_width=16, tx_buffer_enable=False, rx_buffer_enable=False):
        assert data_width in [16, 32]
        # Common signals
        self.data_width     = data_width

        # Control
        self.ready          = Signal() # o

        self.tx_idle        = Signal() # i
        self.tx_polarity    = Signal() # i

        self.tx_cominit_stb = Signal() # i
        self.tx_cominit_ack = Signal() # o
        self.tx_comwake_stb = Signal() # i
        self.tx_comwake_ack = Signal() # o

        self.rx_idle        = Signal() # o
        self.rx_cdrhold     = Signal() # i
        self.rx_polarity    = Signal() # i

        self.rx_cominit_stb = Signal() # o
        self.rx_comwake_stb = Signal() # o

        # Datapath
        self.sink           = stream.Endpoint(phy_description(data_width))
        self.source         = stream.Endpoint(phy_description(data_width))

        # PLL
        self.refclk         = Signal()
        self.cplllock       = Signal()
        self.cpllpd         = Signal()
        self.cpllreset      = Signal()

        # Receive Ports - 8b10b Decoder
        self.rxctrl0        = Signal(data_width//8)
        self.rxctrl1        = Signal(data_width//8)
        self.rxctrl2        = Signal(data_width//8)
        self.rxctrl3        = Signal(data_width//8)

        self.rxcharisk        = Signal(data_width//8)
        self.rxdisperr      = Signal(data_width//8) # o
        self.rxnotintable   = Signal(data_width//8) # o
        rxdisperr           = Signal(data_width//8)

        self.comb += [
            self.rxcharisk.eq(self.rxctrl0),
            rxdisperr.eq(self.rxctrl1)
        ]

        # Receive Ports - RX Data Path interface
        self.rxdata         = Signal(data_width)
        self.rxoutclk       = Signal()
        self.rxusrclk       = Signal()
        self.rxusrclk2      = Signal()

        # Receive Ports - RX Ports for SATA
        self.rxcominitdet   = Signal()
        self.rxcomwakedet   = Signal()

        # Transmit Ports - 8b10b Encoder Control Ports
        self.txcharisk      = Signal(data_width//8)

        # Transmit Ports - TX Data Path interface
        self.txdata         = Signal(data_width)
        self.txoutclk       = Signal()
        self.txusrclk       = Signal()
        self.txusrclk2      = Signal()

        # Transmit Ports - TX Ports for PCI Express
        self.txelecidle     = Signal(reset=1)

        # Transmit Ports - TX Ports for SATA
        self.txcomfinish    = Signal()
        self.txcominit      = Signal()
        self.txcomwake      = Signal()

        # Power-down signals
        self.rxpd           = Signal()
        self.txpd           = Signal()

        # Config at startup
        div_config = {
            "gen1": 4,
            "gen2": 2,
            "gen3": 1,
        }
        rxout_div = div_config[gen]
        txout_div = div_config[gen]

        progdiv = {
            "gen1": 40.0,
            "gen2": 20.0,
            "gen3": 10.0,
        }
        tx_progdiv_cfg = progdiv[gen]
        rx_progdiv_cfg = progdiv[gen]

        # TX Init ----------------------------------------------------------------------------------
        self.submodules.tx_init = tx_init = GTYTXInit(clk_freq, buffer_enable=tx_buffer_enable)
        self.comb += tx_init.plllock.eq(self.cplllock)
        self.comb += self.cpllreset.eq(tx_init.pllreset)

        # RX Init ----------------------------------------------------------------------------------
        self.submodules.rx_init = rx_init = GTYRXInit(clk_freq, buffer_enable=rx_buffer_enable)
        self.comb += rx_init.plllock.eq(self.cplllock)

        # Ready ------------------------------------------------------------------------------------
        self.comb += self.ready.eq(tx_init.done & rx_init.done)

        # Specific / Generic signals encoding/decoding ---------------------------------------------
        self.comb += [
            self.txelecidle.eq(self.tx_idle | self.txpd),
            self.tx_cominit_ack.eq(self.tx_cominit_stb & self.txcomfinish),
            self.tx_comwake_ack.eq(self.tx_comwake_stb & self.txcomfinish),
            self.rx_cominit_stb.eq(self.rxcominitdet),
            self.rx_comwake_stb.eq(self.rxcomwakedet),
        ]
        self.submodules += _RisingEdge(self.tx_cominit_stb, self.txcominit)
        self.submodules += _RisingEdge(self.tx_comwake_stb, self.txcomwake)

        self.sync.sata_rx += [
            self.source.valid.eq(1),
            self.source.charisk.eq(self.rxcharisk),
            self.source.data.eq(self.rxdata)
        ]

        self.sync.sata_tx += [
            self.txcharisk.eq(self.sink.charisk),
            self.txdata.eq(self.sink.data),
            self.sink.ready.eq(1),
        ]

        # Internals and clock domain crossing ------------------------------------------------------
        # sys_clk --> sata_tx clk
        txpd       = Signal()
        txelecidle = Signal(reset=1)
        txcominit  = Signal()
        txcomwake  = Signal()
        self.specials += [
            MultiReg(self.txpd,             txpd, "sata_tx"),
            MultiReg(self.txelecidle, txelecidle, "sata_tx"),
        ]
        self.submodules += [
            _PulseSynchronizer(self.txcominit,  "sys",  txcominit, "sata_tx"),
            _PulseSynchronizer(self.txcomwake,  "sys",  txcomwake, "sata_tx"),
        ]

        # sata_tx clk --> sys clk
        txcomfinish = Signal()
        self.submodules += _PulseSynchronizer(txcomfinish, "sata_tx", self.txcomfinish, "sys")

        # sata_rx clk --> sys clk
        rxcominitdet = Signal()
        rxcomwakedet = Signal()
        rxratedone   = Signal()
        rxdisperr    = Signal(data_width//8)
        rxnotintable = Signal(data_width//8)
        self.specials += [
            MultiReg(rxcominitdet, self.rxcominitdet, "sys"),
            MultiReg(rxcomwakedet, self.rxcomwakedet, "sys"),
            MultiReg(rxdisperr, self.rxdisperr, "sys"),
            MultiReg(rxnotintable, self.rxnotintable, "sys")
        ]

        # DRP mux ----------------------------------------------------------------------------------
        self.drp = DRPInterface()
        self.submodules.drp_mux = drp_mux = DRPMux()
        drp_mux.add_interface(self.drp)

        # GTYE4_CHANNEL instance -------------------------------------------------------------------
        class Open(Signal): pass
        rxphaligndone = Signal()
        self.gty_params = dict(
            p_ACJTAG_DEBUG_MODE              = 0b0,
            p_ACJTAG_MODE                    = 0b0,
            p_ACJTAG_RESET                   = 0b0,
            p_ADAPT_CFG0                     = 0b0000000000000000,
            p_ADAPT_CFG1                     = 0b1111100000011100,
            p_ADAPT_CFG2                     = 0b0000000000000000,
            p_ALIGN_COMMA_DOUBLE             = "FALSE",
            p_ALIGN_COMMA_ENABLE             = 0b1111111111,
            p_ALIGN_COMMA_WORD               = 2,
            p_ALIGN_MCOMMA_DET               = "TRUE",
            p_ALIGN_MCOMMA_VALUE             = 0b1010000011,
            p_ALIGN_PCOMMA_DET               = "TRUE",
            p_ALIGN_PCOMMA_VALUE             = 0b0101111100,
            p_A_RXOSCALRESET                 = 0b0,
            p_A_RXPROGDIVRESET               = 0b0,
            p_A_RXTERMINATION                = 0b1,
            p_A_TXDIFFCTRL                   = 0b01100,
            p_A_TXPROGDIVRESET               = 0b0,
            p_CBCC_DATA_SOURCE_SEL           = "DECODED",
            p_CDR_SWAP_MODE_EN               = 0b0,
            p_CFOK_PWRSVE_EN                 = 0b1,
            p_CHAN_BOND_KEEP_ALIGN           = "FALSE",
            p_CHAN_BOND_MAX_SKEW             = 1,
            p_CHAN_BOND_SEQ_1_1              = 0b0000000000,
            p_CHAN_BOND_SEQ_1_2              = 0b0000000000,
            p_CHAN_BOND_SEQ_1_3              = 0b0000000000,
            p_CHAN_BOND_SEQ_1_4              = 0b0000000000,
            p_CHAN_BOND_SEQ_1_ENABLE         = 0b1111,
            p_CHAN_BOND_SEQ_2_1              = 0b0000000000,
            p_CHAN_BOND_SEQ_2_2              = 0b0000000000,
            p_CHAN_BOND_SEQ_2_3              = 0b0000000000,
            p_CHAN_BOND_SEQ_2_4              = 0b0000000000,
            p_CHAN_BOND_SEQ_2_ENABLE         = 0b1111,
            p_CHAN_BOND_SEQ_2_USE            = "FALSE",
            p_CHAN_BOND_SEQ_LEN              = 1,
            p_CH_HSPMUX                      = 0b0010000000100000,
            p_CKCAL1_CFG_0                   = 0b1100000011000000,
            p_CKCAL1_CFG_1                   = 0b0001000011000000,
            p_CKCAL1_CFG_2                   = 0b0010000000001000,
            p_CKCAL1_CFG_3                   = 0b0000000000000000,
            p_CKCAL2_CFG_0                   = 0b1100000011000000,
            p_CKCAL2_CFG_1                   = 0b1000000011000000,
            p_CKCAL2_CFG_2                   = 0b0001000000000000,
            p_CKCAL2_CFG_3                   = 0b0000000000000000,
            p_CKCAL2_CFG_4                   = 0b0000000000000000,
            p_CLK_CORRECT_USE                = "FALSE",
            p_CLK_COR_KEEP_IDLE              = "FALSE",
            p_CLK_COR_MAX_LAT                = 20 if not rx_buffer_enable else {16:6, 32:12}[data_width],
            p_CLK_COR_MIN_LAT                = 18 if not rx_buffer_enable else {16:4, 32: 8}[data_width],
            p_CLK_COR_PRECEDENCE             = "TRUE",
            p_CLK_COR_REPEAT_WAIT            = 0,
            p_CLK_COR_SEQ_1_1                = 0b0100000000,
            p_CLK_COR_SEQ_1_2                = 0b0100000000,
            p_CLK_COR_SEQ_1_3                = 0b0100000000,
            p_CLK_COR_SEQ_1_4                = 0b0100000000,
            p_CLK_COR_SEQ_1_ENABLE           = 0b1111,
            p_CLK_COR_SEQ_2_1                = 0b0100000000,
            p_CLK_COR_SEQ_2_2                = 0b0100000000,
            p_CLK_COR_SEQ_2_3                = 0b0100000000,
            p_CLK_COR_SEQ_2_4                = 0b0100000000,
            p_CLK_COR_SEQ_2_ENABLE           = 0b1111,
            p_CLK_COR_SEQ_2_USE              = "FALSE",
            p_CLK_COR_SEQ_LEN                = 1,
            p_CPLL_CFG0                      = 0b0000000111111010,
            p_CPLL_CFG1                      = 0b0000000000101011,
            p_CPLL_CFG2                      = 0b0000000000000010,
            p_CPLL_CFG3                      = 0b0000000000000000,
            p_CPLL_FBDIV                     = 5,
            p_CPLL_FBDIV_45                  = 4,
            p_CPLL_INIT_CFG0                 = 0b0000001010110010,
            p_CPLL_LOCK_CFG                  = 0b0000000111101000,
            p_CPLL_REFCLK_DIV                = 1,
            p_CTLE3_OCAP_EXT_CTRL            = 0b000,
            p_CTLE3_OCAP_EXT_EN              = 0b0,
            p_DDI_CTRL                       = 0b00,
            p_DDI_REALIGN_WAIT               = 15,
            p_DEC_MCOMMA_DETECT              = "TRUE",
            p_DEC_PCOMMA_DETECT              = "TRUE",
            p_DEC_VALID_COMMA_ONLY           = "TRUE",
            p_DELAY_ELEC                     = 0b0,
            p_DMONITOR_CFG0                  = 0b0000000000,
            p_DMONITOR_CFG1                  = 0b00000000,
            p_ES_CLK_PHASE_SEL               = 0b0,
            p_ES_CONTROL                     = 0b000000,
            p_ES_ERRDET_EN                   = "FALSE",
            p_ES_EYE_SCAN_EN                 = "FALSE",
            p_ES_HORZ_OFFSET                 = 0b000000000000,
            p_ES_PRESCALE                    = 0b00000,
            p_ES_QUALIFIER0                  = 0b0000000000000000,
            p_ES_QUALIFIER1                  = 0b0000000000000000,
            p_ES_QUALIFIER2                  = 0b0000000000000000,
            p_ES_QUALIFIER3                  = 0b0000000000000000,
            p_ES_QUALIFIER4                  = 0b0000000000000000,
            p_ES_QUALIFIER5                  = 0b0000000000000000,
            p_ES_QUALIFIER6                  = 0b0000000000000000,
            p_ES_QUALIFIER7                  = 0b0000000000000000,
            p_ES_QUALIFIER8                  = 0b0000000000000000,
            p_ES_QUALIFIER9                  = 0b0000000000000000,
            p_ES_QUAL_MASK0                  = 0b0000000000000000,
            p_ES_QUAL_MASK1                  = 0b0000000000000000,
            p_ES_QUAL_MASK2                  = 0b0000000000000000,
            p_ES_QUAL_MASK3                  = 0b0000000000000000,
            p_ES_QUAL_MASK4                  = 0b0000000000000000,
            p_ES_QUAL_MASK5                  = 0b0000000000000000,
            p_ES_QUAL_MASK6                  = 0b0000000000000000,
            p_ES_QUAL_MASK7                  = 0b0000000000000000,
            p_ES_QUAL_MASK8                  = 0b0000000000000000,
            p_ES_QUAL_MASK9                  = 0b0000000000000000,
            p_ES_SDATA_MASK0                 = 0b0000000000000000,
            p_ES_SDATA_MASK1                 = 0b0000000000000000,
            p_ES_SDATA_MASK2                 = 0b0000000000000000,
            p_ES_SDATA_MASK3                 = 0b0000000000000000,
            p_ES_SDATA_MASK4                 = 0b0000000000000000,
            p_ES_SDATA_MASK5                 = 0b0000000000000000,
            p_ES_SDATA_MASK6                 = 0b0000000000000000,
            p_ES_SDATA_MASK7                 = 0b0000000000000000,
            p_ES_SDATA_MASK8                 = 0b0000000000000000,
            p_ES_SDATA_MASK9                 = 0b0000000000000000,
            p_EYESCAN_VP_RANGE               = 0,
            p_EYE_SCAN_SWAP_EN               = 0b0,
            p_FTS_DESKEW_SEQ_ENABLE          = 0b1111,
            p_FTS_LANE_DESKEW_CFG            = 0b1111,
            p_FTS_LANE_DESKEW_EN             = "FALSE",
            p_GEARBOX_MODE                   = 0b00000,
            p_ISCAN_CK_PH_SEL2               = 0b0,
            p_LOCAL_MASTER                   = 0b1,
            p_LPBK_BIAS_CTRL                 = 4,
            p_LPBK_EN_RCAL_B                 = 0b0,
            p_LPBK_EXT_RCAL                  = 0b1000,
            p_LPBK_IND_CTRL0                 = 5,
            p_LPBK_IND_CTRL1                 = 5,
            p_LPBK_IND_CTRL2                 = 5,
            p_LPBK_RG_CTRL                   = 2,
            p_OOBDIVCTL                      = 0b01,
            p_OOB_PWRUP                      = 0b1,
            p_PCI3_AUTO_REALIGN              = "OVR_1K_BLK",
            p_PCI3_PIPE_RX_ELECIDLE          = 0b0,
            p_PCI3_RX_ASYNC_EBUF_BYPASS      = 0b00,
            p_PCI3_RX_ELECIDLE_EI2_ENABLE    = 0b0,
            p_PCI3_RX_ELECIDLE_H2L_COUNT     = 0b000000,
            p_PCI3_RX_ELECIDLE_H2L_DISABLE   = 0b000,
            p_PCI3_RX_ELECIDLE_HI_COUNT      = 0b000000,
            p_PCI3_RX_ELECIDLE_LP4_DISABLE   = 0b0,
            p_PCI3_RX_FIFO_DISABLE           = 0b0,
            p_PCIE3_CLK_COR_EMPTY_THRSH      = 0b00000,
            p_PCIE3_CLK_COR_FULL_THRSH       = 0b010000,
            p_PCIE3_CLK_COR_MAX_LAT          = 0b00100,
            p_PCIE3_CLK_COR_MIN_LAT          = 0b00000,
            p_PCIE3_CLK_COR_THRSH_TIMER      = 0b001000,
            p_PCIE_64B_DYN_CLKSW_DIS         = "FALSE",
            p_PCIE_BUFG_DIV_CTRL             = 0b0001000000000000,
            p_PCIE_GEN4_64BIT_INT_EN         = "FALSE",
            p_PCIE_PLL_SEL_MODE_GEN12        = 0b00,
            p_PCIE_PLL_SEL_MODE_GEN3         = 0b11,
            p_PCIE_PLL_SEL_MODE_GEN4         = 0b10,
            p_PCIE_RXPCS_CFG_GEN3            = 0b0000101010100101,
            p_PCIE_RXPMA_CFG                 = 0b0010100000001010,
            p_PCIE_TXPCS_CFG_GEN3            = 0b0010110010100100,
            p_PCIE_TXPMA_CFG                 = 0b0010100000001010,
            p_PCS_PCIE_EN                    = "FALSE",
            p_PCS_RSVD0                      = 0b0000000000000000,
            p_PD_TRANS_TIME_FROM_P2          = 0b000000111100,
            p_PD_TRANS_TIME_NONE_P2          = 0b00011001,
            p_PD_TRANS_TIME_TO_P2            = 0b01100100,
            p_PREIQ_FREQ_BST                 = 0,
            p_RATE_SW_USE_DRP                = 0b1,
            p_RCLK_SIPO_DLY_ENB              = 0b0,
            p_RCLK_SIPO_INV_EN               = 0b0,
            p_RTX_BUF_CML_CTRL               = 0b011,
            p_RTX_BUF_TERM_CTRL              = 0b00,
            p_RXBUFRESET_TIME                = 0b00011,
            p_RXBUF_ADDR_MODE                = "FAST",
            p_RXBUF_EIDLE_HI_CNT             = 0b1000,
            p_RXBUF_EIDLE_LO_CNT             = 0b0000,
            p_RXBUF_EN                       = "TRUE" if rx_buffer_enable else "FALSE",
            p_RXBUF_RESET_ON_CB_CHANGE       = "TRUE",
            p_RXBUF_RESET_ON_COMMAALIGN      = "FALSE",
            p_RXBUF_RESET_ON_EIDLE           = "FALSE",
            p_RXBUF_RESET_ON_RATE_CHANGE     = "TRUE",
            p_RXBUF_THRESH_OVFLW             = 0 if not rx_buffer_enable else {16:61, 32:57}[data_width],
            p_RXBUF_THRESH_OVRD              = "TRUE" if rx_buffer_enable else "FALSE",
            p_RXBUF_THRESH_UNDFLW            = 4 if not rx_buffer_enable else {16:1, 32:3}[data_width],
        )

        self.gty_params.update(
            p_RXCDRFREQRESET_TIME            = 0b00001,
            p_RXCDRPHRESET_TIME              = 0b00001,
            p_RXCDR_CFG0                     = 0b0000000000000011,
            p_RXCDR_CFG0_GEN3                = 0b0000000000000011,
            p_RXCDR_CFG1                     = 0b0000000000000000,
            p_RXCDR_CFG1_GEN3                = 0b0000000000000000,
            p_RXCDR_CFG2                     = 0b0000001001000010 if gen == "gen1" else
                                               0b0000001001010011 if gen == "gen2" else
                                               0b0000001001100011,
            p_RXCDR_CFG2_GEN2                = 0b1001000010 if gen == "gen1" else
                                               0b1001010011 if gen == "gen2" else
                                               0b1001100011,
            p_RXCDR_CFG2_GEN3                = 0b0000001001000010 if gen == "gen1" else
                                               0b0000001001010011 if gen == "gen2" else
                                               0b0000001001100011,
            p_RXCDR_CFG2_GEN4                = 0b0000000101100100,
            p_RXCDR_CFG3                     = 0b0000000000011010,
            p_RXCDR_CFG3_GEN2                = 0b011010,
            p_RXCDR_CFG3_GEN3                = 0b0000000000011010,
            p_RXCDR_CFG3_GEN4                = 0b0000000000011010,
            p_RXCDR_CFG4                     = 0b0101110011110110,
            p_RXCDR_CFG4_GEN3                = 0b0101110011110110,
            p_RXCDR_CFG5                     = 0b1011010001101011,
            p_RXCDR_CFG5_GEN3                = 0b0001010001101011,
            p_RXCDR_FR_RESET_ON_EIDLE        = 0b0,
            p_RXCDR_HOLD_DURING_EIDLE        = 0b0,
            p_RXCDR_LOCK_CFG0                = 0b0010001000000001,
            p_RXCDR_LOCK_CFG1                = 0b1001111111111111,
            p_RXCDR_LOCK_CFG2                = 0b000000000000000,
            p_RXCDR_LOCK_CFG3                = 0b0000000000000000,
            p_RXCDR_LOCK_CFG4                = 0b0000000000000000,
            p_RXCDR_PH_RESET_ON_EIDLE        = 0b0,
            p_RXCFOK_CFG0                    = 0b0000000000000000,
            p_RXCFOK_CFG1                    = 0b1000000000010101,
            p_RXCFOK_CFG2                    = 0b0000001010101110,
            p_RXCKCAL1_IQ_LOOP_RST_CFG       = 0b0000000000000000,
            p_RXCKCAL1_I_LOOP_RST_CFG        = 0b0000000000000000,
            p_RXCKCAL1_Q_LOOP_RST_CFG        = 0b0000000000000000,
            p_RXCKCAL2_DX_LOOP_RST_CFG       = 0b0000000000000000,
            p_RXCKCAL2_D_LOOP_RST_CFG        = 0b0000000000000000,
            p_RXCKCAL2_S_LOOP_RST_CFG        = 0b0000000000000000,
            p_RXCKCAL2_X_LOOP_RST_CFG        = 0b0000000000000000,
            p_RXDFELPMRESET_TIME             = 0b0001111,
            p_RXDFELPM_KL_CFG0               = 0b000000000000000,
            p_RXDFELPM_KL_CFG1               = 0b1010000010000010,
            p_RXDFELPM_KL_CFG2               = 0b0000000100000000,
            p_RXDFE_CFG0                     = 0b0000101000000000,
            p_RXDFE_CFG1                     = 0b0000000000000000,
            p_RXDFE_GC_CFG0                  = 0b0000000000000000,
            p_RXDFE_GC_CFG1                  = 0b1000000000000000,
            p_RXDFE_GC_CFG2                  = 0b1111111111100000,
            p_RXDFE_H2_CFG0                  = 0b0000000000000000,
            p_RXDFE_H2_CFG1                  = 0b0000000000000010,
            p_RXDFE_H3_CFG0                  = 0b0000000000000000,
            p_RXDFE_H3_CFG1                  = 0b1000000000000010,
            p_RXDFE_H4_CFG0                  = 0b0000000000000000,
            p_RXDFE_H4_CFG1                  = 0b1000000000000010,
            p_RXDFE_H5_CFG0                  = 0b0000000000000000,
            p_RXDFE_H5_CFG1                  = 0b1000000000000010,
            p_RXDFE_H6_CFG0                  = 0b0000000000000000,
            p_RXDFE_H6_CFG1                  = 0b1000000000000010,
            p_RXDFE_H7_CFG0                  = 0b0000000000000000,
            p_RXDFE_H7_CFG1                  = 0b1000000000000010,
            p_RXDFE_H8_CFG0                  = 0b0000000000000000,
            p_RXDFE_H8_CFG1                  = 0b1000000000000010,
            p_RXDFE_H9_CFG0                  = 0b0000000000000000,
            p_RXDFE_H9_CFG1                  = 0b1000000000000010,
            p_RXDFE_HA_CFG0                  = 0b0000000000000000,
            p_RXDFE_HA_CFG1                  = 0b1000000000000010,
            p_RXDFE_HB_CFG0                  = 0b0000000000000000,
            p_RXDFE_HB_CFG1                  = 0b1000000000000010,
            p_RXDFE_HC_CFG0                  = 0b0000000000000000,
            p_RXDFE_HC_CFG1                  = 0b1000000000000010,
            p_RXDFE_HD_CFG0                  = 0b0000000000000000,
            p_RXDFE_HD_CFG1                  = 0b1000000000000010,
            p_RXDFE_HE_CFG0                  = 0b0000000000000000,
            p_RXDFE_HE_CFG1                  = 0b1000000000000010,
            p_RXDFE_HF_CFG0                  = 0b0000000000000000,
            p_RXDFE_HF_CFG1                  = 0b1000000000000010,
            p_RXDFE_KH_CFG0                  = 0b1000000000000000,
            p_RXDFE_KH_CFG1                  = 0b1111111000000000,
            p_RXDFE_KH_CFG2                  = 0b0000001000000000,
            p_RXDFE_KH_CFG3                  = 0b0100000100000001,
            p_RXDFE_OS_CFG0                  = 0b0010000000000000,
            p_RXDFE_OS_CFG1                  = 0b1000000000000000,
            p_RXDFE_UT_CFG0                  = 0b0000000000000000,
            p_RXDFE_UT_CFG1                  = 0b0000000000000011,
            p_RXDFE_UT_CFG2                  = 0b0000000000000000,
            p_RXDFE_VP_CFG0                  = 0b0000000000000000,
            p_RXDFE_VP_CFG1                  = 0b0000000000110011,
            p_RXDLY_CFG                      = 0b0000000000010000,
            p_RXDLY_LCFG                     = 0b0000000000110000,
            p_RXELECIDLE_CFG                 = "SIGCFG_4",
            p_RXGBOX_FIFO_INIT_RD_ADDR       = 4,
            p_RXGEARBOX_EN                   = "FALSE",
            p_RXISCANRESET_TIME              = 0b00001,
            p_RXLPM_CFG                      = 0b0000000000000000,
            p_RXLPM_GC_CFG                   = 0b1111100000000000,
            p_RXLPM_KH_CFG0                  = 0b0000000000000000,
            p_RXLPM_KH_CFG1                  = 0b1010000000000010,
            p_RXLPM_OS_CFG0                  = 0b0000000000000000,
            p_RXLPM_OS_CFG1                  = 0b1000000000000010,
            p_RXOOB_CFG                      = 0b000000110,
            p_RXOOB_CLK_CFG                  = "PMA",
            p_RXOSCALRESET_TIME              = 0b00011,
            p_RXOUT_DIV                      = rxout_div,
            p_RXPCSRESET_TIME                = 0b00011,
            p_RXPHBEACON_CFG                 = 0b0000000000000000,
            p_RXPHDLY_CFG                    = 0b0010000001110000,
            p_RXPHSAMP_CFG                   = 0b0010000100000000,
            p_RXPHSLIP_CFG                   = 0b1001100100110011,
            p_RXPH_MONITOR_SEL               = 0b00000,
            p_RXPI_CFG0                      = 0b0000001100000001,
            p_RXPI_CFG1                      = 0b0000000011111100,
            p_RXPMACLK_SEL                   = "DATA",
            p_RXPMARESET_TIME                = 0b00011,
            p_RXPRBS_ERR_LOOPBACK            = 0b0,
            p_RXPRBS_LINKACQ_CNT             = 15,
            p_RXREFCLKDIV2_SEL               = 0b0,
            p_RXSLIDE_AUTO_WAIT              = 7,
            p_RXSLIDE_MODE                   = "PCS",
            p_RXSYNC_MULTILANE               = 0b0,
            p_RXSYNC_OVRD                    = 0b0,
            p_RXSYNC_SKIP_DA                 = 0b0,
            p_RX_AFE_CM_EN                   = 0b0,
            p_RX_BIAS_CFG0                   = 0b0001001010110000,
            p_RX_BUFFER_CFG                  = 0b000000,
            p_RX_CAPFF_SARC_ENB              = 0b0,
            p_RX_CLK25_DIV                   = 6,
            p_RX_CLKMUX_EN                   = 0b1,
            p_RX_CLK_SLIP_OVRD               = 0b00000,
            p_RX_CM_BUF_CFG                  = 0b1010,
            p_RX_CM_BUF_PD                   = 0b0,
            p_RX_CM_SEL                      = 3,
            p_RX_CM_TRIM                     = 10,
            p_RX_CTLE_PWR_SAVING             = 0b0,
            p_RX_CTLE_RES_CTRL               = 0b0000,
            p_RX_DATA_WIDTH                  = 20 if data_width == 16 else 40,
            p_RX_DDI_SEL                     = 0b000000,
            p_RX_DEFER_RESET_BUF_EN          = "TRUE",
            p_RX_DEGEN_CTRL                  = 0b100,
            p_RX_DFELPM_CFG0                 = 10,
            p_RX_DFELPM_CFG1                 = 0b1,
            p_RX_DFELPM_KLKH_AGC_STUP_EN     = 0b1,
            p_RX_DFE_AGC_CFG1                = 2,
            p_RX_DFE_KL_LPM_KH_CFG0          = 3,
            p_RX_DFE_KL_LPM_KH_CFG1          = 2,
            p_RX_DFE_KL_LPM_KL_CFG0          = 0b11,
            p_RX_DFE_KL_LPM_KL_CFG1          = 2,
            p_RX_DFE_LPM_HOLD_DURING_EIDLE   = 0b0,
            p_RX_DISPERR_SEQ_MATCH           = "TRUE",
            p_RX_DIVRESET_TIME               = 0b00001,
            p_RX_EN_CTLE_RCAL_B              = 0b0,
            p_RX_EN_SUM_RCAL_B               = 0,
            p_RX_EYESCAN_VS_CODE             = 0b0000000,
            p_RX_EYESCAN_VS_NEG_DIR          = 0b0,
            p_RX_EYESCAN_VS_RANGE            = 0b10,
            p_RX_EYESCAN_VS_UT_SIGN          = 0b0,
            p_RX_FABINT_USRCLK_FLOP          = 0b0,
            p_RX_I2V_FILTER_EN               = 0b1,
            p_RX_INT_DATAWIDTH               = 0 if data_width == 16 else 1,
            p_RX_PMA_POWER_SAVE              = 0b0,
            p_RX_PMA_RSV0                    = 0b0000000000101111,
            p_RX_PROGDIV_CFG                 = rx_progdiv_cfg,
            p_RX_PROGDIV_RATE                = 0b0000000000000001,
            p_RX_RESLOAD_CTRL                = 0b0000,
            p_RX_RESLOAD_OVRD                = 0b0,
            p_RX_SAMPLE_PERIOD               = 0b111,
            p_RX_SIG_VALID_DLY               = 11,
            p_RX_SUM_DEGEN_AVTT_OVERITE      = 0,
            p_RX_SUM_DFETAPREP_EN            = 0b0,
            p_RX_SUM_IREF_TUNE               = 0b0000,
            p_RX_SUM_PWR_SAVING              = 0,
            p_RX_SUM_RES_CTRL                = 0b0000,
            p_RX_SUM_VCMTUNE                 = 0b0011,
            p_RX_SUM_VCM_BIAS_TUNE_EN        = 0b1,
            p_RX_SUM_VCM_OVWR                = 0b0,
            p_RX_SUM_VREF_TUNE               = 0b100,
            p_RX_TUNE_AFE_OS                 = 0b10,
            p_RX_VREG_CTRL                   = 0b010,
            p_RX_VREG_PDB                    = 0b1,
            p_RX_WIDEMODE_CDR                = 0b00,
            p_RX_WIDEMODE_CDR_GEN3           = 0b00,
            p_RX_WIDEMODE_CDR_GEN4           = 0b01,
            p_RX_XCLK_SEL                    = "RXDES" if rx_buffer_enable else "RXUSR",
            p_RX_XMODE_SEL                   = 0b1,
            p_SAMPLE_CLK_PHASE               = 0b0,
            p_SAS_12G_MODE                   = 0b0,
            p_SATA_BURST_SEQ_LEN             = 0b0110,
            p_SATA_BURST_VAL                 = 0b100,
            p_SATA_CPLL_CFG                  = "VCO_3000MHZ",
            p_SATA_EIDLE_VAL                 = 0b100,
            p_SHOW_REALIGN_COMMA             = "FALSE",
            p_SIM_DEVICE                     = "ULTRASCALE_PLUS",
            p_SIM_MODE                       = "FAST",
            p_SIM_RECEIVER_DETECT_PASS       = "TRUE",
            p_SIM_RESET_SPEEDUP              = "TRUE",
            p_SIM_TX_EIDLE_DRIVE_LEVEL       = "Z",
            p_SRSTMODE                       = 0b0,
            p_TAPDLY_SET_TX                  = 0b00,
            p_TERM_RCAL_CFG                  = 0b100001000000010,
            p_TERM_RCAL_OVRD                 = 0b001,
            p_TRANS_TIME_RATE                = 0b00001110,
            p_TST_RSV0                       = 0b00000000,
            p_TST_RSV1                       = 0b00000000,
        )

        self.gty_params.update(
            p_TXBUF_EN                       = "TRUE" if tx_buffer_enable else "FALSE",
            p_TXBUF_RESET_ON_RATE_CHANGE     = "TRUE",
            p_TXDLY_CFG                      = 0b1000000000010000,
            p_TXDLY_LCFG                     = 0b0000000000110000,
            p_TXDRV_FREQBAND                 = 0,
            p_TXFE_CFG0                      = 0b0000001111000010,
            p_TXFE_CFG1                      = 0b0110110000000000,
            p_TXFE_CFG2                      = 0b0110110000000000,
            p_TXFE_CFG3                      = 0b0110110000000000,
            p_TXFIFO_ADDR_CFG                = "LOW",
            p_TXGBOX_FIFO_INIT_RD_ADDR       = 4,
            p_TXGEARBOX_EN                   = "FALSE",
            p_TXOUT_DIV                      = txout_div,
            p_TXPCSRESET_TIME                = 0b00011,
            p_TXPHDLY_CFG0                   = 0b0110000001110000,
            p_TXPHDLY_CFG1                   = 0b0000000000001010 if gen == "gen3" else
                                               0b0000000000001111,
            p_TXPH_CFG                       = 0b0000011100100011 if gen == "gen1" else
                                               0b0000001100100011,
            p_TXPH_CFG2                      = 0b0000000000000000,
            p_TXPH_MONITOR_SEL               = 0b00000,
            p_TXPI_CFG0                      = 0b0000001100000000,
            p_TXPI_CFG1                      = 0b0111010101010101,
            p_TXPI_GRAY_SEL                  = 0b0,
            p_TXPI_INVSTROBE_SEL             = 0b0,
            p_TXPI_PPM                       = 0b0,
            p_TXPI_PPM_CFG                   = 0b00000000,
            p_TXPI_SYNFREQ_PPM               = 0b001,
            p_TXPMARESET_TIME                = 0b00011,
            p_TXREFCLKDIV2_SEL               = 0b0,
            p_TXSWBST_BST                    = 1,
            p_TXSWBST_EN                     = 0,
            p_TXSWBST_MAG                    = 4,
            p_TXSYNC_MULTILANE               = 0b0,
            p_TXSYNC_OVRD                    = 0b0,
            p_TXSYNC_SKIP_DA                 = 0b0,
            p_TX_CLK25_DIV                   = 6,
            p_TX_CLKMUX_EN                   = 0b1,
            p_TX_DATA_WIDTH                  = 20 if data_width == 16 else 40,
            p_TX_DCC_LOOP_RST_CFG            = 0b0000000000000100,
            p_TX_DEEMPH0                     = 0b000000,
            p_TX_DEEMPH1                     = 0b000000,
            p_TX_DEEMPH2                     = 0b000000,
            p_TX_DEEMPH3                     = 0b000000,
            p_TX_DIVRESET_TIME               = 0b00001,
            p_TX_DRIVE_MODE                  = "DIRECT",
            p_TX_EIDLE_ASSERT_DELAY          = 0b100,
            p_TX_EIDLE_DEASSERT_DELAY        = 0b011,
            p_TX_FABINT_USRCLK_FLOP          = 0b0,
            p_TX_FIFO_BYP_EN                 = 0 if tx_buffer_enable else 1,
            p_TX_IDLE_DATA_ZERO              = 0b0,
            p_TX_INT_DATAWIDTH               = 0 if data_width == 16 else 1,
            p_TX_LOOPBACK_DRIVE_HIZ          = "FALSE",
            p_TX_MAINCURSOR_SEL              = 0b0,
            p_TX_MARGIN_FULL_0               = 0b1011000,
            p_TX_MARGIN_FULL_1               = 0b1010111,
            p_TX_MARGIN_FULL_2               = 0b1010101,
            p_TX_MARGIN_FULL_3               = 0b1010011,
            p_TX_MARGIN_FULL_4               = 0b1010001,
            p_TX_MARGIN_LOW_0                = 0b1001100,
            p_TX_MARGIN_LOW_1                = 0b1001011,
            p_TX_MARGIN_LOW_2                = 0b1001000,
            p_TX_MARGIN_LOW_3                = 0b1000010,
            p_TX_MARGIN_LOW_4                = 0b1000000,
            p_TX_PHICAL_CFG0                 = 0b0000000000100000,
            p_TX_PHICAL_CFG1                 = 0b0000000001000000,
            p_TX_PI_BIASSET                  = 0,
            p_TX_PMADATA_OPT                 = 0b0,
            p_TX_PMA_POWER_SAVE              = 0b0,
            p_TX_PMA_RSV0                    = 0b0000000000000000,
            p_TX_PMA_RSV1                    = 0b0000000000000000,
            p_TX_PROGCLK_SEL                 = "PREPI",
            p_TX_PROGDIV_CFG                 = tx_progdiv_cfg,
            p_TX_PROGDIV_RATE                = 0b0000000000000001,
            p_TX_RXDETECT_CFG                = 0b00000000110010,
            p_TX_RXDETECT_REF                = 5,
            p_TX_SAMPLE_PERIOD               = 0b111,
            p_TX_SW_MEAS                     = 0b00,
            p_TX_VREG_CTRL                   = 0b011,
            p_TX_VREG_PDB                    = 0b1,
            p_TX_VREG_VREFSEL                = 0b10,
            p_TX_XCLK_SEL                    = "TXOUT" if tx_buffer_enable else "TXUSR",
            p_USB_BOTH_BURST_IDLE            = 0b0,
            p_USB_BURSTMAX_U3WAKE            = 0b1111111,
            p_USB_BURSTMIN_U3WAKE            = 0b1100011,
            p_USB_CLK_COR_EQ_EN              = 0b0,
            p_USB_EXT_CNTL                   = 0b1,
            p_USB_IDLEMAX_POLLING            = 0b1010111011,
            p_USB_IDLEMIN_POLLING            = 0b0100101011,
            p_USB_LFPSPING_BURST             = 0b000000101,
            p_USB_LFPSPOLLING_BURST          = 0b000110001,
            p_USB_LFPSPOLLING_IDLE_MS        = 0b000000100,
            p_USB_LFPSU1EXIT_BURST           = 0b000011101,
            p_USB_LFPSU2LPEXIT_BURST_MS      = 0b001100011,
            p_USB_LFPSU3WAKE_BURST_MS        = 0b111110011,
            p_USB_LFPS_TPERIOD               = 0b0011,
            p_USB_LFPS_TPERIOD_ACCURATE      = 0b1,
            p_USB_MODE                       = 0b0,
            p_USB_PCIE_ERR_REP_DIS           = 0b0,
            p_USB_PING_SATA_MAX_INIT         = 21,
            p_USB_PING_SATA_MIN_INIT         = 12,
            p_USB_POLL_SATA_MAX_BURST        = 8,
            p_USB_POLL_SATA_MIN_BURST        = 4,
            p_USB_RAW_ELEC                   = 0b0,
            p_USB_RXIDLE_P0_CTRL             = 0b1,
            p_USB_TXIDLE_TUNE_ENABLE         = 0b1,
            p_USB_U1_SATA_MAX_WAKE           = 7,
            p_USB_U1_SATA_MIN_WAKE           = 4,
            p_USB_U2_SAS_MAX_COM             = 64,
            p_USB_U2_SAS_MIN_COM             = 36,
            p_USE_PCS_CLK_PHASE_SEL          = 0b0,
            p_Y_ALL_MODE                     = 0b0
        )
        self.gty_params.update(
            #
            i_PCSRSVDIN       = 0x00,
            i_GTRSVD          = 0x0000,
            i_TSTIN           = 0, #2**20-1,

            # Reset modes
            i_GTTXRESETSEL    = 0,
            i_GTRXRESETSEL    = 0,
            i_RESETOVRD       = 0,

            # DRP
            i_DRPADDR         = drp_mux.addr,
            i_DRPCLK          = drp_mux.clk,
            i_DRPDI           = drp_mux.di,
            o_DRPDO           = drp_mux.do,
            i_DRPEN           = drp_mux.en,
            o_DRPRDY          = drp_mux.rdy,
            i_DRPWE           = drp_mux.we,

            # CPLL
            i_CPLLRESET       = 0,
            i_CPLLPD          = self.cpllreset,
            o_CPLLLOCK        = self.cplllock,
            i_CPLLLOCKEN      = 1,
            i_CPLLREFCLKSEL   = 0b111,
            i_GTGREFCLK       = self.refclk,

            # QPLL
            i_QPLL0CLK        = 0,
            i_QPLL0REFCLK     = 0,
            i_QPLL1CLK        = 0,
            i_QPLL1REFCLK     = 0,
            i_QPLL0FREQLOCK   = 0,
            i_QPLL1FREQLOCK   = 0,

            # TX clock
            o_TXOUTCLK        = self.txoutclk,
            i_TXSYSCLKSEL     = 0b00,
            i_TXPLLCLKSEL     = 0b00,
            i_TXOUTCLKSEL     = 0b010 if tx_buffer_enable else 0b101,

            # TX Startup/Reset
            i_GTTXRESET       = tx_init.gtXxreset,
            o_TXRESETDONE     = tx_init.Xxresetdone,
            i_TXDLYSRESET     = tx_init.Xxdlysreset,
            o_TXDLYSRESETDONE = tx_init.Xxdlysresetdone,
            o_TXPHALIGNDONE   = tx_init.Xxphaligndone,
            i_TXUSERRDY       = tx_init.Xxuserrdy,
            i_TXSYNCMODE      = 1,
            i_TXDLYBYPASS     = 1 if tx_buffer_enable else 0,
            i_TXPHDLYPD       = 1 if tx_buffer_enable else 0,

            # Transmit Ports - TX 8B/10B Encoder Ports
            i_TX8B10BBYPASS   = 0,

            # TX data
            i_TXCTRL0         = 0,
            i_TXCTRL1         = 0,
            i_TXCTRL2         = self.txcharisk,
            i_TXDATA          = self.txdata,
            i_TXUSRCLK        = self.txusrclk,
            i_TXUSRCLK2       = self.txusrclk2,

            # Power-Down Ports
            i_RXPD             = Replicate(self.rxpd, 2),
            i_TXPD             = Replicate(txpd, 2),

            # TX electrical
            i_TXDIFFCTRL      = 0b1100,
            i_TXINHIBIT       = 0,

            # Transmit Ports - PCI Express Ports
            i_TXPDELECIDLEMODE= 0,
            i_TXELECIDLE      = txelecidle,
            i_TXMARGIN        = 0b000,
            i_TXRATE          = 0b000,
            i_TXSWING         = 0,

            # Internal Loopback
            i_LOOPBACK        = 0b000,

            # RX Startup/Reset
            i_GTRXRESET       = rx_init.gtXxreset,
            o_RXRESETDONE     = rx_init.Xxresetdone,
            i_RXDLYSRESET     = rx_init.Xxdlysreset,
            o_RXDLYSRESETDONE = rx_init.Xxdlysresetdone,
            i_RXPMARESET      = 0,
            o_RXPHALIGNDONE   = rxphaligndone,
            i_RXSYNCALLIN     = rxphaligndone,
            i_RXUSERRDY       = rx_init.Xxuserrdy,
            i_RXSYNCIN        = 0,
            i_RXSYNCMODE      = 1,
            o_RXSYNCDONE      = rx_init.Xxsyncdone,
            i_RXDLYBYPASS     = 1 if rx_buffer_enable else 0,
            i_RXPHDLYPD       = 1 if rx_buffer_enable else 0,

            # RX AFE
            i_RXDFEXYDEN      = 1,
            i_RXLPMEN         = 1,

            # RX clock
            i_RXRATE          = 0,
            i_RXSYSCLKSEL     = 0b00,
            i_RXOUTCLKSEL     = 0b010,
            i_RXPLLCLKSEL     = 0b00,
            o_RXOUTCLK        = self.rxoutclk,
            i_RXUSRCLK        = self.rxusrclk,
            i_RXUSRCLK2       = self.rxusrclk2,

            # FPGA RX Interface Datapath Configuration
            i_RX8B10BEN       = 1,

            # Receive Ports - CDR Ports
            i_RXCDRFREQRESET  = 0,
            i_RXCDRHOLD       = self.rx_cdrhold,
            o_RXCDRLOCK       = Open(),
            i_RXCDROVRDEN     = 0,
            i_RXCDRRESET      = 0,
            i_RXOSCALRESET    = 0,
            o_RXOSINTDONE     = Open(),
            o_RXOSINTSTARTED  = Open(),
            o_RXOSINTSTROBEDONE    = Open(),
            o_RXOSINTSTROBESTARTED = Open(),

            # Receive Ports - RX Byte and Word Alignment Ports
            o_RXBYTEISALIGNED = Open(),
            o_RXBYTEREALIGN   = Open(),
            o_RXCOMMADET      = Open(),
            i_RXCOMMADETEN    = 1,
            i_RXMCOMMAALIGNEN = 1,
            i_RXPCOMMAALIGNEN = 1,
            i_RXSLIDE         = 0,

            o_GTPOWERGOOD     = Open(),

            # Receive Ports - RX OOB Signaling ports
            o_RXCOMSASDET     = Open(),
            o_RXCOMWAKEDET    = rxcomwakedet,
            o_RXCOMINITDET    = rxcominitdet,

            # Receive Ports - RX OOB signalling Ports
            o_RXELECIDLE      = Open(),
            i_RXELECIDLEMODE  = 0b00,

            # Transmit Ports - TX OOB signaling Ports
            o_TXCOMFINISH     = txcomfinish,
            i_TXCOMINIT       = txcominit,
            i_TXCOMSAS        = 0,
            i_TXCOMWAKE       = txcomwake,

            # FPGA TX Interface Datapath Configuration
            i_TX8B10BEN       = 1,

            # RX data
            o_RXCTRL0         = self.rxctrl0,
            o_RXCTRL1         = self.rxctrl1,
            o_RXCTRL2         = self.rxctrl2,
            o_RXCTRL3         = self.rxctrl3,
            o_RXDATA          = self.rxdata,

            # Polarity
            i_TXPOLARITY      = self.tx_polarity,
            i_RXPOLARITY      = self.rx_polarity,

            # Pads
            i_GTYRXP          = pads.rx_p,
            i_GTYRXN          = pads.rx_n,
            o_GTYTXP          = pads.tx_p,
            o_GTYTXN          = pads.tx_n
        )

        self.specials += Instance("GTYE4_CHANNEL", **self.gty_params)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2016 Olof Kindgren <olof.kindgren@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *

# Layouts ------------------------------------------------------------------------------------------

tx_to_rx = [
    ("write",    1),
    ("read",     1),
    ("identify", 1),
    ("count",    16)
]

rx_to_tx = [
    ("dma_activate", 1),
    ("d2h_error",    1)
]

# LiteSATA Command TX ------------------------------------------------------------------------------

class LiteSATACommandTX(Module):
    def __init__(self, transport):
        self.sink    = sink    = stream.Endpoint(command_tx_description(32))
        self.to_rx   = to_rx   = stream.Endpoint(tx_to_rx)
        self.from_rx = from_rx = stream.Endpoint(rx_to_tx)

        # # #

        self.comb += [
            transport.sink.pm_port.eq(0),
            transport.sink.features.eq(0),
            transport.sink.lba.eq(sink.sector),
            transport.sink.device.eq(0xe0),
            transport.sink.count.eq(sink.count),
            transport.sink.icc.eq(0),
            transport.sink.control.eq(0),
            transport.sink.data.eq(sink.data)
        ]

        is_write       = Signal()
        is_read        = Signal()
        is_identify    = Signal()
        dwords_counter = Signal(max=fis_max_dwords)

        self.fsm = fsm = FSM(reset_state="IDLE")
        self.submodules += fsm
        fsm.act("IDLE",
            sink.ready.eq(0),
            If(sink.valid,
                NextState("SEND_CMD")
            ).Else(
                sink.ready.eq(1)
            )
        )
        self.sync += \
            If(fsm.ongoing("IDLE"),
                is_write.eq(sink.write),
                is_read.eq(sink.read),
                is_identify.eq(sink.identify),
            )

        fsm.act("SEND_CMD",
            transport.sink.valid.eq(sink.valid),
            transport.sink.last.eq(1),
            transport.sink.c.eq(1),
            If(transport.sink.valid & transport.sink.ready,
                If(is_write,
                    NextState("WAIT_DMA_ACTIVATE")
                ).Else(
                    sink.ready.eq(1),
                    NextState("IDLE")
                )
            )
        )
        fsm.act("WAIT_DMA_ACTIVATE",
            NextValue(dwords_counter, 0),
            If(from_rx.dma_activate,
                NextState("SEND_DATA")
            ).Elif(from_rx.d2h_error,
                sink.ready.eq(1),
                NextState("IDLE")
            )
        )
        fsm.act("SEND_DATA",
            transport.sink.valid.eq(sink.valid),
            transport.sink.last.eq((dwords_counter == (fis_max_dwords-1)) | sink.last),
            sink.ready.eq(transport.sink.ready),
            If(sink.valid & sink.ready,
                NextValue(dwords_counter, dwords_counter + 1),
                If(sink.last,
                    NextState("IDLE")
                ).Elif(dwords_counter == (fis_max_dwords-1),
                    NextState("WAIT_DMA_ACTIVATE")
                )
            )
        )
        self.comb += \
            If(fsm.ongoing("SEND_DATA"),
                transport.sink.type.eq(fis_types["DATA"]),
            ).Else(
                transport.sink.type.eq(fis_types["REG_H2D"]),
                If(is_write,
                    transport.sink.command.eq(regs["WRITE_DMA_EXT"])
                ).Elif(is_read,
                    transport.sink.command.eq(regs["READ_DMA_EXT"]),
                ).Else(
                    transport.sink.command.eq(regs["IDENTIFY_DEVICE"]),
                )
            )
        self.comb += [
            If(sink.valid,
                to_rx.write.eq(sink.write),
                to_rx.read.eq(sink.read),
                to_rx.identify.eq(sink.identify),
                to_rx.count.eq(sink.count)
            )
        ]

# LiteSATA Command RX ------------------------------------------------------------------------------

class LiteSATACommandRX(Module):
    def __init__(self, transport):
        self.source  = source  = stream.Endpoint(command_rx_description(32))
        self.to_tx   = to_tx   = stream.Endpoint(rx_to_tx)
        self.from_tx = from_tx = stream.Endpoint(tx_to_rx)

        # Debug
        self.d2h_status = Signal(8)
        self.d2h_errors = Signal(8)

        # # #

        def test_type(name):
            return transport.source.type == fis_types[name]

        is_identify     = Signal()
        is_dma_activate = Signal()
        read_ndwords    = Signal(max=sectors2dwords(2**16))
        dwords_counter  = Signal(max=sectors2dwords(2**16))
        read_done       = Signal()

        self.sync += \
            If(from_tx.read,
                read_ndwords.eq(from_tx.count*sectors2dwords(1) - 1)
            )
        self.comb += read_done.eq(dwords_counter == read_ndwords)

        d2h_error     = Signal()
        clr_d2h_error = Signal()
        set_d2h_error = Signal()
        self.sync += \
            If(clr_d2h_error,
                d2h_error.eq(0)
            ).Elif(set_d2h_error,
                d2h_error.eq(1)
            )

        read_error     = Signal()
        clr_read_error = Signal()
        set_read_error = Signal()
        self.sync += \
            If(clr_read_error,
                read_error.eq(0)
            ).Elif(set_read_error,
                read_error.eq(1)
            )

        update_d2h = Signal()
        self.sync += \
            If(update_d2h,
                self.d2h_status.eq(transport.source.status),
                self.d2h_errors.eq(transport.source.errors)
            )

        self.fsm = fsm = FSM(reset_state="IDLE")
        self.submodules += fsm
        fsm.act("IDLE",
            NextValue(dwords_counter, 0),
            transport.source.ready.eq(1),
            clr_d2h_error.eq(1),
            clr_read_error.eq(1),
            If(from_tx.write,
                NextState("WAIT_WRITE_ACTIVATE_OR_REG_D2H")
            ).Elif(from_tx.read,
                NextState("WAIT_READ_DATA_OR_REG_D2H"),
            ).Elif(from_tx.identify,
                NextState("WAIT_PIO_SETUP_D2H"),
            )
        )
        self.sync += \
            If(fsm.ongoing("IDLE"),
                is_identify.eq(from_tx.identify)
            )
        fsm.act("WAIT_WRITE_ACTIVATE_OR_REG_D2H",
            transport.source.ready.eq(1),
            If(transport.source.valid,
                If(test_type("DMA_ACTIVATE_D2H"),
                    is_dma_activate.eq(1),
                ).Elif(test_type("REG_D2H"),
                    update_d2h.eq(1),
                    set_d2h_error.eq(transport.source.status[reg_d2h_status["err"]]),
                    NextState("PRESENT_WRITE_RESPONSE")
                )
            )
        )
        fsm.act("PRESENT_WRITE_RESPONSE",
            source.valid.eq(1),
            source.last.eq(1),
            source.write.eq(1),
            source.end.eq(1),
            source.failed.eq(transport.source.error | d2h_error),
            If(source.valid & source.ready,
                NextState("IDLE")
            )
        )
        fsm.act("WAIT_READ_DATA_OR_REG_D2H",
            transport.source.ready.eq(1),
            If(transport.source.valid,
                transport.source.ready.eq(0),
                If(test_type("DATA"),
                    NextState("PRESENT_READ_DATA")
                ).Elif(test_type("REG_D2H"),
                    update_d2h.eq(1),
                    set_d2h_error.eq(transport.source.status[reg_d2h_status["err"]]),
                    NextState("PRESENT_READ_RESPONSE")
                )
            )
        )
        fsm.act("WAIT_PIO_SETUP_D2H",
            transport.source.ready.eq(1),
            If(transport.source.valid,
                transport.source.ready.eq(0),
                If(test_type("PIO_SETUP_D2H"),
                    NextState("PRESENT_PIO_SETUP_D2H")
                ).Else(
                    NextState("FLUSH")
                )
            )
        )
        fsm.act("PRESENT_PIO_SETUP_D2H",
            transport.source.ready.eq(1),
            If(transport.source.valid & transport.source.last,
                NextState("WAIT_READ_DATA_OR_REG_D2H")
            )
        )

        fsm.act("PRESENT_READ_DATA",
            set_read_error.eq(transport.source.error),
            source.valid.eq(transport.source.valid),
            source.last.eq(transport.source.last),
            source.read.eq(~is_identify),
            source.identify.eq(is_identify),
            source.failed.eq(transport.source.error),
            source.end.eq(is_identify),
            source.data.eq(transport.source.data),
            transport.source.ready.eq(source.ready),
            If(source.valid & source.ready,
                If(~read_done, NextValue(dwords_counter, dwords_counter + 1)),
                If(source.last,
                    If(is_identify,
                        NextState("IDLE")
                    ).Else(
                        NextState("WAIT_READ_DATA_OR_REG_D2H")
                    )
                )
            )
        )

        fsm.act("PRESENT_READ_RESPONSE",
            source.valid.eq(1),
            source.last.eq(1),
            source.read.eq(1),
            source.end.eq(1),
            source.failed.eq(~read_done | read_error | d2h_error),
            If(source.valid & source.ready,
                NextState("IDLE")
            )
        )

        fsm.act("FLUSH",
            transport.source.ready.eq(1),
            If(transport.source.valid & transport.source.last,
               NextState("WAIT_PIO_SETUP_D2H")
            )
        )
        self.comb += [
            to_tx.dma_activate.eq(is_dma_activate),
            to_tx.d2h_error.eq(d2h_error)
        ]

# LiteSATA Command ---------------------------------------------------------------------------------

class LiteSATACommand(Module):
    def __init__(self, transport):
        self.submodules.tx = LiteSATACommandTX(transport)
        self.submodules.rx = LiteSATACommandRX(transport)
        self.comb += [
            self.rx.to_tx.connect(self.tx.from_rx),
            self.tx.to_rx.connect(self.rx.from_tx)
        ]
        self.sink, self.source = self.tx.sink, self.rx.source



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2016 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from litesata.common import *
from litesata.core.link import LiteSATALink

from litex.soc.interconnect.stream_sim import *

from test.model.hdd import *


class LinkStreamer(PacketStreamer):
    def __init__(self):
        PacketStreamer.__init__(self, link_description(32), packet_cls=LinkTXPacket)


class LinkLogger(PacketLogger):
    def __init__(self):
        PacketLogger.__init__(self, link_description(32), packet_cls=LinkRXPacket)


class TestLink(unittest.TestCase):
    def test_link(self):
        def generator(dut):
            for i in range(2):
                streamer_packet = LinkTXPacket([i for i in range(64)])
                yield from dut.streamer.send_blocking(streamer_packet)
                yield from dut.logger.receive()

                # check results
                s, l, e = check(streamer_packet, dut.logger.packet)
                print("shift " + str(s) + " / length " + str(l) + " / errors " + str(e))
                self.assertEqual(s, 0)
                self.assertEqual(e, 0)

        class DUT(Module):
            def __init__(self):
                self.submodules.hdd = HDD(
                        link_debug         = False,
                        link_random_level  = 50,
                        transport_debug    = False,
                        transport_loopback = True)
                self.submodules.link = LiteSATALink(self.hdd.phy)

                self.submodules.streamer = LinkStreamer()
                self.submodules.streamer_randomizer = Randomizer(link_description(32), level=50)

                self.submodules.logger_randomizer = Randomizer(link_description(32), level=50)
                self.submodules.logger = LinkLogger()

                self.submodules.pipeline = Pipeline(
                    self.streamer,
                    self.streamer_randomizer,
                    self.link,
                    self.logger_randomizer,
                    self.logger
                )

        dut = DUT()
        generators = {
            "sys" :   [generator(dut),
                       dut.hdd.link.generator(),
                       dut.streamer.generator(),
                       dut.streamer_randomizer.generator(),
                       dut.logger.generator(),
                       dut.logger_randomizer.generator(),
                       dut.hdd.phy.rx.generator(),
                       dut.hdd.phy.tx.generator()]
        }
        clocks = {"sys": 10}
        run_simulation(dut, generators, clocks)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *
from litesata.core.link import LiteSATALink
from litesata.core.transport import LiteSATATransport
from litesata.core.command import LiteSATACommand

# LiteSATA Core ------------------------------------------------------------------------------------

class LiteSATACore(Module):
    def __init__(self, phy):
        self.submodules.link      = LiteSATALink(phy)
        self.submodules.transport = LiteSATATransport(self.link)
        self.submodules.command   = LiteSATACommand(self.transport)
        self.sink, self.source = self.command.sink, self.command.source



#
# This file is part of LiteSATA.
#
# Copyright (c) 2019-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *
from litesata.common import _PulseSynchronizer, _RisingEdge

from migen.genlib.cdc import MultiReg
from migen.genlib.resetsync import AsyncResetSynchronizer

from litex.soc.cores.clock import S7MMCM

from liteiclink.serdes.gtp_7series_init import GTPTXInit, GTPRXInit
from liteiclink.serdes.gtp_7series import GTPQuadPLL

# --------------------------------------------------------------------------------------------------

class A7LiteSATAPHYCRG(Module):
    def __init__(self, refclk, pads, gtp, gen, tx_buffer_enable=False):
        self.tx_reset = Signal()
        self.rx_reset = Signal()

        self.clock_domains.cd_sata_tx = ClockDomain()
        self.clock_domains.cd_sata_rx = ClockDomain()

        # CPLL -------------------------------------------------------------------------------------
        #   (gen3) 150MHz / VCO @ 3GHz / Linerate @ 6Gbps
        #   (gen2 & gen1) VCO still @ 3 GHz, Linerate is decreased with output dividers.
        if isinstance(refclk, (Signal, ClockSignal)):
            self.refclk = refclk
        else:
            self.refclk = Signal()
            self.specials += Instance("IBUFDS_GTE2",
                i_CEB = 0,
                i_I   = pads.clk_p,
                i_IB  = pads.clk_n,
                o_O   = self.refclk
            )
        self.comb += gtp.gtrefclk0.eq(self.refclk)

        # TX clocking ------------------------------------------------------------------------------
        #   (gen3) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 300MHz (16-bits) /  150MHz (32-bits)
        #   (gen2) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 150MHz (16-bits) /   75MHz (32-bits)
        #   (gen1) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 75MHz  (16-bits) / 37.5MHz (32-bits)
        if tx_buffer_enable:
            self.specials += Instance("BUFG", i_I=gtp.txoutclk, o_O=self.cd_sata_tx.clk)
        else:
            tx_mmcm_clkin = Signal()
            tx_mmcm = S7MMCM(speedgrade=-1)
            tx_mmcm_clkout_freq = {
                "gen1":  75e6/(gtp.data_width/16),
                "gen2": 150e6/(gtp.data_width/16),
                "gen3": 300e6/(gtp.data_width/16),
            }
            self.submodules += tx_mmcm
            self.specials += Instance("BUFG", i_I=gtp.txoutclk, o_O=tx_mmcm_clkin)
            tx_mmcm.register_clkin(tx_mmcm_clkin, 150e6)
            tx_mmcm.create_clkout(self.cd_sata_tx, tx_mmcm_clkout_freq[gen], with_reset=False)

        self.comb += gtp.txusrclk.eq(self.cd_sata_tx.clk)
        self.comb += gtp.txusrclk2.eq(self.cd_sata_tx.clk)

        # RX clocking ------------------------------------------------------------------------------
        #   (gen3) sata_rx recovered clk @ 300MHz (16-bits) /  150MHz (32-bits) from GTP RXOUTCLK
        #   (gen2) sata_rx recovered clk @ 150MHz (16-bits) /   75MHz (32-bits) from GTP RXOUTCLK
        #   (gen1) sata_rx recovered clk @ 75MHz  (16-bits) / 37.5MHz (32-bits) from GTP RXOUTCLK
        self.specials += Instance("BUFG", i_I=gtp.rxoutclk, o_O=self.cd_sata_rx.clk)

        self.comb += gtp.rxusrclk.eq(self.cd_sata_rx.clk)
        self.comb += gtp.rxusrclk2.eq(self.cd_sata_rx.clk)

        # Reset for SATA TX/RX clock domains -------------------------------------------------------
        self.specials += [
            AsyncResetSynchronizer(self.cd_sata_tx, ~gtp.qplllock | self.tx_reset),
            AsyncResetSynchronizer(self.cd_sata_rx, ~gtp.qplllock | self.rx_reset)
        ]

# --------------------------------------------------------------------------------------------------

class A7LiteSATAPHY(Module):
    def __init__(self, pads, gen, clk_freq, data_width=16, tx_buffer_enable=False, rx_buffer_enable=False, qpll=None):
        assert data_width in [16, 32]
        # Common signals
        self.data_width     = data_width

        # Control
        self.ready          = Signal() # o

        self.tx_idle        = Signal() # i
        self.tx_polarity    = Signal() # i

        self.tx_cominit_stb = Signal() # i
        self.tx_cominit_ack = Signal() # o
        self.tx_comwake_stb = Signal() # i
        self.tx_comwake_ack = Signal() # o

        self.rx_idle        = Signal() # o
        self.rx_cdrhold     = Signal() # i
        self.rx_polarity    = Signal() # i

        self.rx_cominit_stb = Signal() # o
        self.rx_comwake_stb = Signal() # o

        self.rxdisperr      = Signal(data_width//8) # o
        self.rxnotintable   = Signal(data_width//8) # o

        # Datapath
        self.sink           = stream.Endpoint(phy_description(data_width))
        self.source         = stream.Endpoint(phy_description(data_width))

        # K7 specific signals
        # Channel - Ref Clock Ports
        self.gtrefclk0      = Signal()

        # Quadl PLL
        self.qplllock       = Signal()

        # Receive Ports - 8b10b Decoder
        self.rxcharisk      = Signal(data_width//8)

        # Receive Ports - RX Data Path interface
        self.rxdata         = Signal(data_width)
        self.rxoutclk       = Signal()
        self.rxusrclk       = Signal()
        self.rxusrclk2      = Signal()

        # Receive Ports - RX Ports for SATA
        self.rxcominitdet   = Signal()
        self.rxcomwakedet   = Signal()

        # Transmit Ports - 8b10b Encoder Control Ports
        self.txcharisk      = Signal(data_width//8)

        # Transmit Ports - TX Data Path interface
        self.txdata         = Signal(data_width)
        self.txoutclk       = Signal()
        self.txusrclk       = Signal()
        self.txusrclk2      = Signal()

        # Transmit Ports - TX Ports for PCI Express
        self.txelecidle     = Signal(reset=1)

        # Transmit Ports - TX Ports for SATA
        self.txcomfinish    = Signal()
        self.txcominit      = Signal()
        self.txcomwake      = Signal()

        # Power-down signals
        self.cpllpd         = Signal()
        self.rxpd           = Signal()
        self.txpd           = Signal()

        # Config at startup
        div_config = {
            "gen1": 4,
            "gen2": 2,
            "gen3": 1,
        }
        rxout_div = div_config[gen]
        txout_div = div_config[gen]

        cdr_config = {
            "gen1": 0x0000047fe106024481010,
            "gen2": 0x0000047fe206024481010,
            "gen3": 0x0000087fe206024441010,
        }
        rxcdr_cfg = cdr_config[gen]

        # TX Init ----------------------------------------------------------------------------------
        self.submodules.tx_init = tx_init = GTPTXInit(clk_freq, buffer_enable=tx_buffer_enable)
        self.comb += tx_init.plllock.eq(self.qplllock)

        # RX Init ----------------------------------------------------------------------------------
        self.submodules.rx_init = rx_init = GTPRXInit(clk_freq, buffer_enable=rx_buffer_enable)
        self.comb += rx_init.plllock.eq(self.qplllock)

        # Ready ------------------------------------------------------------------------------------
        self.comb += self.ready.eq(tx_init.done & rx_init.done)

        # Specific / Generic signals encoding/decoding ---------------------------------------------
        self.comb += [
            self.txelecidle.eq(self.tx_idle | self.txpd),
            self.tx_cominit_ack.eq(self.tx_cominit_stb & self.txcomfinish),
            self.tx_comwake_ack.eq(self.tx_comwake_stb & self.txcomfinish),
            self.rx_cominit_stb.eq(self.rxcominitdet),
            self.rx_comwake_stb.eq(self.rxcomwakedet),
        ]
        self.submodules += _RisingEdge(self.tx_cominit_stb, self.txcominit)
        self.submodules += _RisingEdge(self.tx_comwake_stb, self.txcomwake)

        self.sync.sata_rx += [
            self.source.valid.eq(1),
            self.source.charisk.eq(self.rxcharisk),
            self.source.data.eq(self.rxdata)
        ]

        self.sync.sata_tx += [
            self.txcharisk.eq(self.sink.charisk),
            self.txdata.eq(self.sink.data),
            self.sink.ready.eq(1),
        ]

        # Internals and clock domain crossing ------------------------------------------------------
        # sys_clk --> sata_tx clk
        txpd       = Signal()
        txelecidle = Signal(reset=1)
        txcominit  = Signal()
        txcomwake  = Signal()
        self.specials += [
            MultiReg(self.txpd,             txpd, "sata_tx"),
            MultiReg(self.txelecidle, txelecidle, "sata_tx"),
        ]
        self.submodules += [
            _PulseSynchronizer(self.txcominit,  "sys",  txcominit, "sata_tx"),
            _PulseSynchronizer(self.txcomwake,  "sys",  txcomwake, "sata_tx"),
        ]

        # sata_tx clk --> sys clk
        txcomfinish = Signal()
        self.submodules += _PulseSynchronizer(txcomfinish, "sata_tx", self.txcomfinish, "sys")

        # sata_rx clk --> sys clk
        rxcominitdet = Signal()
        rxcomwakedet = Signal()
        rxratedone   = Signal()
        rxdisperr    = Signal(data_width//8)
        rxnotintable = Signal(data_width//8)
        self.specials += [
            MultiReg(rxcominitdet, self.rxcominitdet, "sys"),
            MultiReg(rxcomwakedet, self.rxcomwakedet, "sys"),
            MultiReg(rxdisperr, self.rxdisperr, "sys"),
            MultiReg(rxnotintable, self.rxnotintable, "sys")
        ]

        # QPLL ------------------------------------------------------------------------------------
        linerate_config = {
            "gen1": 1.5e9,
            "gen2": 3.0e9,
            "gen3": 6.0e9
        }
        if qpll is None:
            self.submodules.qpll = qpll = GTPQuadPLL(self.gtrefclk0, 150e6, linerate_config[gen])
            qpll.index = 0
        self.comb += self.qplllock.eq(qpll.lock)
        self.comb += qpll.reset.eq(tx_init.pllreset)

        # OOB clock (75MHz) ------------------------------------------------------------------------
        oobclk = Signal()
        self.specials += Instance("FDPE",
            p_INIT = 1,
            i_CE   = 1,
            i_PRE  = 0,
            i_C    = self.gtrefclk0,
            i_D    = ~oobclk,
            o_Q    = oobclk
        )

        # GTPE2_CHANNEL Instance -------------------------------------------------------------------
        class Open(Signal): pass
        rxphaligndone = Signal()
        gtp_params = dict(
            # Simulation-Only Attributes
            p_SIM_RECEIVER_DETECT_PASS   = "TRUE",
            p_SIM_TX_EIDLE_DRIVE_LEVEL   = "X",
            p_SIM_RESET_SPEEDUP          = "FALSE",
            p_SIM_VERSION                = "2.0",

            # RX Byte and Word Alignment Attributes
            p_ALIGN_COMMA_DOUBLE         = "FALSE",
            p_ALIGN_COMMA_ENABLE         = 0b1111111111,
            p_ALIGN_COMMA_WORD           = 2 if data_width == 16 else 4,
            p_ALIGN_MCOMMA_DET           = "TRUE",
            p_ALIGN_MCOMMA_VALUE         = 0b1010000011,
            p_ALIGN_PCOMMA_DET           = "TRUE",
            p_ALIGN_PCOMMA_VALUE         = 0b0101111100,
            p_SHOW_REALIGN_COMMA         = "TRUE",
            p_RXSLIDE_AUTO_WAIT          = 7,
            p_RXSLIDE_MODE               = "OFF" if rx_buffer_enable else "PCS",
            p_RX_SIG_VALID_DLY           = 10,

            # RX 8B/10B Decoder Attributes
            p_RX_DISPERR_SEQ_MATCH       = "TRUE",
            p_DEC_MCOMMA_DETECT          = "TRUE",
            p_DEC_PCOMMA_DETECT          = "TRUE",
            p_DEC_VALID_COMMA_ONLY       = "TRUE",

            # RX Clock Correction Attributes
            p_CBCC_DATA_SOURCE_SEL       = "DECODED",
            p_CLK_COR_SEQ_2_USE          = "FALSE",
            p_CLK_COR_KEEP_IDLE          = "FALSE",
            p_CLK_COR_MAX_LAT            = 9 if data_width == 16 else 20,
            p_CLK_COR_MIN_LAT            = 7 if data_width == 16 else 16,
            p_CLK_COR_PRECEDENCE         = "TRUE",
            p_CLK_COR_REPEAT_WAIT        = 0,
            p_CLK_COR_SEQ_LEN            = 1,
            p_CLK_COR_SEQ_1_ENABLE       = 0b1111,
            p_CLK_COR_SEQ_1_1            = 0b0100000000,
            p_CLK_COR_SEQ_1_2            = 0b0000000000,
            p_CLK_COR_SEQ_1_3            = 0b0000000000,
            p_CLK_COR_SEQ_1_4            = 0b0000000000,
            p_CLK_CORRECT_USE            = "FALSE",
            p_CLK_COR_SEQ_2_ENABLE       = 0b1111,
            p_CLK_COR_SEQ_2_1            = 0b0100000000,
            p_CLK_COR_SEQ_2_2            = 0b0000000000,
            p_CLK_COR_SEQ_2_3            = 0b0000000000,
            p_CLK_COR_SEQ_2_4            = 0b0000000000,

            # RX Channel Bonding Attributes
            p_CHAN_BOND_KEEP_ALIGN       = "FALSE",
            p_CHAN_BOND_MAX_SKEW         = 1,
            p_CHAN_BOND_SEQ_LEN          = 1,
            p_CHAN_BOND_SEQ_1_1          = 0b0000000000,
            p_CHAN_BOND_SEQ_1_2          = 0b0000000000,
            p_CHAN_BOND_SEQ_1_3          = 0b0000000000,
            p_CHAN_BOND_SEQ_1_4          = 0b0000000000,
            p_CHAN_BOND_SEQ_1_ENABLE     = 0b1111,
            p_CHAN_BOND_SEQ_2_1          = 0b0000000000,
            p_CHAN_BOND_SEQ_2_2          = 0b0000000000,
            p_CHAN_BOND_SEQ_2_3          = 0b0000000000,
            p_CHAN_BOND_SEQ_2_4          = 0b0000000000,
            p_CHAN_BOND_SEQ_2_ENABLE     = 0b1111,
            p_CHAN_BOND_SEQ_2_USE        = "FALSE",
            p_FTS_DESKEW_SEQ_ENABLE      = 0b1111,
            p_FTS_LANE_DESKEW_CFG        = 0b1111,
            p_FTS_LANE_DESKEW_EN         = "FALSE",

            # RX Margin Analysis Attributes
            p_ES_CONTROL                 = 0b000000,
            p_ES_ERRDET_EN               = "FALSE",
            p_ES_EYE_SCAN_EN             = "TRUE",
            p_ES_HORZ_OFFSET             = 0x010,
            p_ES_PMA_CFG                 = 0b0000000000,
            p_ES_PRESCALE                = 0b00000,
            p_ES_QUALIFIER               = 0x00000000000000000000,
            p_ES_QUAL_MASK               = 0x00000000000000000000,
            p_ES_SDATA_MASK              = 0x00000000000000000000,
            p_ES_VERT_OFFSET             = 0b000000000,

            # FPGA RX Interface Attributes
            p_RX_DATA_WIDTH              = 20 if data_width == 16 else 40,

            # PMA Attributes
            p_OUTREFCLK_SEL_INV          = 0b11,
            p_PMA_RSV                    = 0x00000333,
            p_PMA_RSV2                   = 0x00002040,
            p_PMA_RSV3                   = 0b00,
            p_PMA_RSV4                   = 0b0000,
            p_RX_BIAS_CFG                = 0b0000111100110011,
            p_DMONITOR_CFG               = 0x000A00,
            p_RX_CM_SEL                  = 0b11,
            p_RX_CM_TRIM                 = 0b1010,
            p_RX_DEBUG_CFG               = 0b00000000000000,
            p_RX_OS_CFG                  = 0b0000010000000,
            p_TERM_RCAL_CFG              = 0b100001000010000,
            p_TERM_RCAL_OVRD             = 0b000,
            p_TST_RSV                    = 0x00000000,
            p_RX_CLK25_DIV               = 6,
            p_TX_CLK25_DIV               = 6,
            p_UCODEER_CLR                = 0b0,

            # PCI Express Attributes
            p_PCS_PCIE_EN                = "FALSE",

            # PCS Attributes
            p_PCS_RSVD_ATTR              = 0x100,

            # RX Buffer Attributes
            p_RXBUF_ADDR_MODE            = "FAST",
            p_RXBUF_EIDLE_HI_CNT         = 0b1000,
            p_RXBUF_EIDLE_LO_CNT         = 0b0000,
            p_RXBUF_EN                   = "TRUE" if rx_buffer_enable else "FALSE",
            p_RX_BUFFER_CFG              = 0b000000,
            p_RXBUF_RESET_ON_CB_CHANGE   = "TRUE",
            p_RXBUF_RESET_ON_COMMAALIGN  = "FALSE",
            p_RXBUF_RESET_ON_EIDLE       = "FALSE",
            p_RXBUF_RESET_ON_RATE_CHANGE = "TRUE",
            p_RXBUFRESET_TIME            = 0b00001,
            p_RXBUF_THRESH_OVFLW         = 61,
            p_RXBUF_THRESH_OVRD          = "FALSE",
            p_RXBUF_THRESH_UNDFLW        = 4,
            p_RXDLY_CFG                  = 0x001F,
            p_RXDLY_LCFG                 = 0x030,
            p_RXDLY_TAP_CFG              = 0x0000,
            p_RXPH_CFG                   = 0xC00002,
            p_RXPHDLY_CFG                = 0x084020,
            p_RXPH_MONITOR_SEL           = 0b00000,
            p_RX_XCLK_SEL                = "RXREC" if rx_buffer_enable else "RXUSR",
            p_RX_DDI_SEL                 = 0b000000,
            p_RX_DEFER_RESET_BUF_EN      = "TRUE",

            # CDR Attributes
            p_RXCDR_CFG                  = rxcdr_cfg,
            p_RXCDR_FR_RESET_ON_EIDLE    = 0b0,
            p_RXCDR_HOLD_DURING_EIDLE    = 0b0,
            p_RXCDR_PH_RESET_ON_EIDLE    = 0b0,
            p_RXCDR_LOCK_CFG             = 0b001001,

            # RX Initialization and Reset Attributes
            p_RXCDRFREQRESET_TIME        = 0b00001,
            p_RXCDRPHRESET_TIME          = 0b00001,
            p_RXISCANRESET_TIME          = 0b00001,
            p_RXPCSRESET_TIME            = 0b00001,
            p_RXPMARESET_TIME            = 0b00011,

            # RX OOB Signaling Attributes
            p_RXOOB_CFG                  = 0b0000110,

            # RX Gearbox Attributes
            p_RXGEARBOX_EN               = "FALSE",
            p_GEARBOX_MODE               = 0b000,

            # PRBS Detection Attribute
            p_RXPRBS_ERR_LOOPBACK        = 0b0,

            # Power-Down Attributes
            p_PD_TRANS_TIME_FROM_P2      = 0x03c,
            p_PD_TRANS_TIME_NONE_P2      = 0x3c,
            p_PD_TRANS_TIME_TO_P2        = 0x64,

            # RX OOB Signaling Attributes
            p_SAS_MAX_COM                = 64,
            p_SAS_MIN_COM                = 36,
            p_SATA_BURST_SEQ_LEN         = 0b0101,
            p_SATA_BURST_VAL             = 0b100,
            p_SATA_EIDLE_VAL             = 0b100,
            p_SATA_MAX_BURST             = 8,
            p_SATA_MAX_INIT              = 21,
            p_SATA_MAX_WAKE              = 7,
            p_SATA_MIN_BURST             = 4,
            p_SATA_MIN_INIT              = 12,
            p_SATA_MIN_WAKE              = 4,

            # RX Fabric Clock Output Control Attributes
            p_TRANS_TIME_RATE            = 0x0E,

            # TX Buffer Attributes
            p_TXBUF_EN                   = "TRUE" if tx_buffer_enable else "FALSE",
            p_TXBUF_RESET_ON_RATE_CHANGE = "TRUE",
            p_TXDLY_CFG                  = 0x001F,
            p_TXDLY_LCFG                 = 0x030,
            p_TXDLY_TAP_CFG              = 0x0000,
            p_TXPH_CFG                   = 0x0780,
            p_TXPHDLY_CFG                = 0x084020,
            p_TXPH_MONITOR_SEL           = 0b00000,
            p_TX_XCLK_SEL                = "TXOUT" if tx_buffer_enable else "TXUSR",

            # FPGA TX Interface Attributes
            p_TX_DATA_WIDTH              = 20 if data_width == 16 else 40,

            # TX Configurable Driver Attributes
            p_TX_DEEMPH0                 = 0b00000,
            p_TX_DEEMPH1                 = 0b00000,
            p_TX_EIDLE_ASSERT_DELAY      = 0b110,
            p_TX_EIDLE_DEASSERT_DELAY    = 0b100,
            p_TX_LOOPBACK_DRIVE_HIZ      = "FALSE",
            p_TX_MAINCURSOR_SEL          = 0b0,
            p_TX_DRIVE_MODE              = "DIRECT",
            p_TX_MARGIN_FULL_0           = 0b1001110,
            p_TX_MARGIN_FULL_1           = 0b1001001,
            p_TX_MARGIN_FULL_2           = 0b1000101,
            p_TX_MARGIN_FULL_3           = 0b1000010,
            p_TX_MARGIN_FULL_4           = 0b1000000,
            p_TX_MARGIN_LOW_0            = 0b1000110,
            p_TX_MARGIN_LOW_1            = 0b1000100,
            p_TX_MARGIN_LOW_2            = 0b1000010,
            p_TX_MARGIN_LOW_3            = 0b1000000,
            p_TX_MARGIN_LOW_4            = 0b1000000,

            # TX Gearbox Attributes
            p_TXGEARBOX_EN               = "FALSE",

            # TX Initialization and Reset Attributes
            p_TXPCSRESET_TIME            = 0b00001,
            p_TXPMARESET_TIME            = 0b00001,

            # TX Receiver Detection Attributes
            p_TX_RXDETECT_CFG            = 0x1832,
            p_TX_RXDETECT_REF            = 0b100,

            # JTAG Attributes
            p_ACJTAG_DEBUG_MODE          = 0b0,
            p_ACJTAG_MODE                = 0b0,
            p_ACJTAG_RESET               = 0b0,

            # CDR Attributes
            p_CFOK_CFG                   = 0x49000040E80,
            p_CFOK_CFG2                  = 0b0100000,
            p_CFOK_CFG3                  = 0b0100000,
            p_CFOK_CFG4                  = 0b0,
            p_CFOK_CFG5                  = 0x0,
            p_CFOK_CFG6                  = 0b0000,
            p_RXOSCALRESET_TIME          = 0b00011,
            p_RXOSCALRESET_TIMEOUT       = 0b00000,

            # PMA Attributes
            p_CLK_COMMON_SWING           = 0b0,
            p_RX_CLKMUX_EN               = 0b1,
            p_TX_CLKMUX_EN               = 0b1,
            p_ES_CLK_PHASE_SEL           = 0b0,
            p_USE_PCS_CLK_PHASE_SEL      = 0b0,
            p_PMA_RSV6                   = 0b0,
            p_PMA_RSV7                   = 0b0,

            # TX Configuration Driver Attributes
            p_TX_PREDRIVER_MODE          = 0b0,
            p_PMA_RSV5                   = 0b0,
            p_SATA_PLL_CFG               = "VCO_3000MHZ",

            # RX Fabric Clock Output Control Attributes
            p_RXOUT_DIV                  = rxout_div,

            # TX Fabric Clock Output Control Attributes
            p_TXOUT_DIV                  = txout_div,

            # RX Phase Interpolator Attributes
            p_RXPI_CFG0                  = 0b000,
            p_RXPI_CFG1                  = 0b1,
            p_RXPI_CFG2                  = 0b1,

            # RX Equalizer Attributes
            p_ADAPT_CFG0                 = 0x00000,
            p_RXLPMRESET_TIME            = 0b0001111,
            p_RXLPM_BIAS_STARTUP_DISABLE = 0b0,
            p_RXLPM_CFG                  = 0b0110,
            p_RXLPM_CFG1                 = 0b0,
            p_RXLPM_CM_CFG               = 0b0,
            p_RXLPM_GC_CFG               = 0b111100010,
            p_RXLPM_GC_CFG2              = 0b001,
            p_RXLPM_HF_CFG               = 0b00001111110000,
            p_RXLPM_HF_CFG2              = 0b01010,
            p_RXLPM_HF_CFG3              = 0b0000,
            p_RXLPM_HOLD_DURING_EIDLE    = 0b0,
            p_RXLPM_INCM_CFG             = 0b1,
            p_RXLPM_IPCM_CFG             = 0b0,
            p_RXLPM_LF_CFG               = 0b000000001111110000,
            p_RXLPM_LF_CFG2              = 0b01010,
            p_RXLPM_OSINT_CFG            = 0b100,

            # TX Phase Interpolator PPM Controller Attributes
            p_TXPI_CFG0                  = 0b00,
            p_TXPI_CFG1                  = 0b00,
            p_TXPI_CFG2                  = 0b00,
            p_TXPI_CFG3                  = 0b0,
            p_TXPI_CFG4                  = 0b0,
            p_TXPI_CFG5                  = 0b000,
            p_TXPI_GREY_SEL              = 0b0,
            p_TXPI_INVSTROBE_SEL         = 0b0,
            p_TXPI_PPMCLK_SEL            = "TXUSRCLK2",
            p_TXPI_PPM_CFG               = 0x00,
            p_TXPI_SYNFREQ_PPM           = 0b001,

            # LOOPBACK Attributes
            p_LOOPBACK_CFG               = 0b0,
            p_PMA_LOOPBACK_CFG           = 0b0,

            # RX OOB Signalling Attributes
            p_RXOOB_CLK_CFG              = "PMA",

            # TX OOB Signalling Attributes
            p_TXOOB_CFG                  = 0b0,

            # RX Buffer Attributes
            p_RXSYNC_MULTILANE           = 0b0,
            p_RXSYNC_OVRD                = 0b0,
            p_RXSYNC_SKIP_DA             = 0b0,

            # TX Buffer Attributes
            p_TXSYNC_MULTILANE           = 0b0,
            p_TXSYNC_OVRD                = 0b0 if tx_buffer_enable else 0b1,
            p_TXSYNC_SKIP_DA             = 0b0
        )
        gtp_params.update(
            # CPLL Ports
            i_GTRSVD               = 0b0000000000000000,
            i_PCSRSVDIN            = 0b0000000000000000,
            i_TSTIN                = 0b11111111111111111111,

            # Channel - DRP Ports
            i_DRPADDR              = rx_init.drp.addr,
            i_DRPCLK               = rx_init.drp.clk,
            i_DRPDI                = rx_init.drp.di,
            o_DRPDO                = rx_init.drp.do,
            i_DRPEN                = rx_init.drp.en,
            o_DRPRDY               = rx_init.drp.rdy,
            i_DRPWE                = rx_init.drp.we,

            # Clocking Ports
            i_RXSYSCLKSEL          = {0: 0b00, 1: 0b11}[qpll.index],
            i_TXSYSCLKSEL          = {0: 0b00, 1: 0b11}[qpll.index],

            # FPGA TX Interface Datapath Configuration
            i_TX8B10BEN            = 1,

            # GTPE2_CHANNEL Clocking Ports
            i_PLL0CLK              = {0: qpll.clk,    1: 0}[qpll.index],
            i_PLL0REFCLK           = {0: qpll.refclk, 1: 0}[qpll.index],
            i_PLL1CLK              = {0: 0, 1: qpll.clk,  }[qpll.index],
            i_PLL1REFCLK           = {0: 0, 1: qpll.refclk}[qpll.index],

            # Loopback Ports
            i_LOOPBACK             = 0b000,

            # PCI Express Ports
            o_PHYSTATUS            = Open(),
            i_RXRATE               = 0,
            o_RXVALID              = Open(),

            # PMA Reserved Ports
            i_PMARSVDIN3           = 0b0,
            i_PMARSVDIN4           = 0b0,

            # Power-Down Ports
            i_RXPD                 = Replicate(self.rxpd, 2),
            i_TXPD                 = Replicate(txpd, 2),

            # RX 8B/10B Decoder Ports
            i_SETERRSTATUS         = 0,

            # RX Initialization and Reset Ports
            i_EYESCANRESET         = 0,
            i_RXUSERRDY            = rx_init.rxuserrdy,

            # RX Margin Analysis Ports
            o_EYESCANDATAERROR     = Open(),
            i_EYESCANMODE          = 0,
            i_EYESCANTRIGGER       = 0,

            # Receive Ports
            i_CLKRSVD0             = 0,
            i_CLKRSVD1             = 0,
            i_DMONFIFORESET        = 0,
            i_DMONITORCLK          = 0,
            o_RXPMARESETDONE       = rx_init.rxpmaresetdone,
            i_SIGVALIDCLK          = oobclk,

            # Receive Ports - CDR Ports
            i_RXCDRFREQRESET       = 0,
            i_RXCDRHOLD            = self.rx_cdrhold,
            o_RXCDRLOCK            = Open(),
            i_RXCDROVRDEN          = 0,
            i_RXCDRRESET           = 0,
            i_RXCDRRESETRSV        = 0,
            i_RXOSCALRESET         = 0,
            i_RXOSINTCFG           = 0b0010,
            o_RXOSINTDONE          = Open(),
            i_RXOSINTHOLD          = 0,
            i_RXOSINTOVRDEN        = 0,
            i_RXOSINTPD            = 0,
            i_RXOSINTSTROBE        = 0,
            o_RXOSINTSTROBESTARTED = Open(),
            i_RXOSINTTESTOVRDEN    = 0,

            # Receive Ports - Clock Correction Ports
            o_RXCLKCORCNT          = Open(),

            # Receive Ports - FPGA RX Interface Datapath Configuration
            i_RX8B10BEN            = 1,

            # Receive Ports - FPGA RX Interface Ports
            i_RXUSRCLK             = self.rxusrclk,
            i_RXUSRCLK2            = self.rxusrclk2,

            # Receive Ports - FPGA RX interface Ports
            o_RXDATA               = self.rxdata,
            # Receive Ports - Pattern Checker Ports
            o_RXPRBSERR            = Open(),
            i_RXPRBSSEL            = 0b000,

            # Receive Ports - Pattern Checker ports
            i_RXPRBSCNTRESET       = 0,

            # Receive Ports - RX 8B/10B Decoder Ports
            o_RXCHARISCOMMA        = Open(),
            o_RXCHARISK            = self.rxcharisk,
            o_RXDISPERR            = rxdisperr,
            o_RXNOTINTABLE         = rxnotintable,

            # Receive Ports - RX AFE Ports
            i_GTPRXN               = pads.rx_n,
            i_GTPRXP               = pads.rx_p,
            i_PMARSVDIN2           = 0b0,
            o_PMARSVDOUT0          = Open(),
            o_PMARSVDOUT1          = Open(),

            # Receive Ports - RX Buffer Bypass Ports
            i_RXBUFRESET           = 0,
            o_RXBUFSTATUS          = Open(),
            i_RXDDIEN              = 0 if rx_buffer_enable else 1,
            i_RXDLYBYPASS          = 1 if rx_buffer_enable else 0,
            i_RXDLYEN              = 0,
            i_RXDLYOVRDEN          = 0,
            i_RXDLYSRESET          = rx_init.rxdlysreset,
            o_RXDLYSRESETDONE      = rx_init.rxdlysresetdone,
            i_RXPHALIGN            = 0,
            o_RXPHALIGNDONE        = rxphaligndone,
            i_RXPHALIGNEN          = 0,
            i_RXPHDLYPD            = 0,
            i_RXPHDLYRESET         = 0,
            o_RXPHMONITOR          = Open(),
            i_RXPHOVRDEN           = 0,
            o_RXPHSLIPMONITOR      = Open(),
            o_RXSTATUS             = Open(),
            i_RXSYNCALLIN          = rxphaligndone,
            o_RXSYNCDONE           = rx_init.rxsyncdone,
            i_RXSYNCIN             = 0,
            i_RXSYNCMODE           = 0 if rx_buffer_enable else 1,
            o_RXSYNCOUT            = Open(),

            # Receive Ports - RX Byte and Word Alignment Ports
            o_RXBYTEISALIGNED      = Open(),
            o_RXBYTEREALIGN        = Open(),
            o_RXCOMMADET           = Open(),
            i_RXCOMMADETEN         = 1,
            i_RXMCOMMAALIGNEN      = 1,
            i_RXPCOMMAALIGNEN      = 1,
            i_RXSLIDE              = 0,

            # Receive Ports - RX Channel Bonding Ports
            o_RXCHANBONDSEQ        = Open(),
            i_RXCHBONDEN           = 0,
            i_RXCHBONDI            = 0b0000,
            i_RXCHBONDLEVEL        = 0,
            i_RXCHBONDMASTER       = 0,
            o_RXCHBONDO            = Open(),
            i_RXCHBONDSLAVE        = 0,

            # Receive Ports - RX Channel Bonding Ports
            o_RXCHANISALIGNED      = Open(),
            o_RXCHANREALIGN        = Open(),

            # Receive Ports - RX Decision Feedback Equalizer
            o_DMONITOROUT          = Open(),
            i_RXADAPTSELTEST       = 0,
            i_RXDFEXYDEN           = 0,
            i_RXOSINTEN            = 0b1,
            i_RXOSINTID0           = 0,
            i_RXOSINTNTRLEN        = 0,
            o_RXOSINTSTROBEDONE    = Open(),

            # Receive Ports - RX Driver,OOB signalling,Coupling and Eq.,CDR
            i_RXLPMLFOVRDEN        = 0,
            i_RXLPMOSINTNTRLEN     = 0,

            # Receive Ports - RX Equalizer Ports
            i_RXLPMHFHOLD          = 0,
            i_RXLPMHFOVRDEN        = 0,
            i_RXLPMLFHOLD          = 0,
            i_RXOSHOLD             = 0,
            i_RXOSOVRDEN           = 0,

            # Receive Ports - RX Fabric ClocK Output Control Ports
            o_RXRATEDONE           = Open(),

            # Receive Ports - RX Fabric Clock Output Control Ports
            i_RXRATEMODE           = 0b0,

            # Receive Ports - RX Fabric Output Control Ports
            o_RXOUTCLK             = self.rxoutclk,
            o_RXOUTCLKFABRIC       = Open(),
            o_RXOUTCLKPCS          = Open(),
            i_RXOUTCLKSEL          = 0b010,

            # Receive Ports - RX Gearbox Ports
            o_RXDATAVALID          = Open(),
            o_RXHEADER             = Open(),
            o_RXHEADERVALID        = Open(),
            o_RXSTARTOFSEQ         = Open(),
            i_RXGEARBOXSLIP        = 0,

            # Receive Ports - RX Initialization and Reset Ports
            i_GTRXRESET            = rx_init.gtrxreset,
            i_RXLPMRESET           = 0,
            i_RXOOBRESET           = 0,
            i_RXPCSRESET           = 0,
            i_RXPMARESET           = 0,

            # Receive Ports - RX OOB Signaling ports
            o_RXCOMSASDET          = Open(),
            o_RXCOMWAKEDET         = rxcomwakedet,

            # Receive Ports - RX OOB Signaling ports
            o_RXCOMINITDET         = rxcominitdet,

            # Receive Ports - RX OOB signalling Ports
            o_RXELECIDLE           = Open(),
            i_RXELECIDLEMODE       = 0b00,

            # Receive Ports - RX Polarity Control Ports
            i_RXPOLARITY           = self.rx_polarity,

            # Receive Ports -RX Initialization and Reset Ports
            o_RXRESETDONE          = rx_init.rxresetdone,

            # TX Buffer Bypass Ports
            i_TXPHDLYTSTCLK        = 0,

            # TX Configurable Driver Ports
            i_TXPOSTCURSOR         = 0b00000,
            i_TXPOSTCURSORINV      = 0,
            i_TXPRECURSOR          = 0b00000,
            i_TXPRECURSORINV       = 0,

            # TX Fabric Clock Output Control Ports
            i_TXRATEMODE           = 0,

            # TX Initialization and Reset Ports
            i_CFGRESET             = 0,
            i_GTTXRESET            = tx_init.gttxreset,
            o_PCSRSVDOUT           = Open(),
            i_TXUSERRDY            = tx_init.txuserrdy,

            # TX Phase Interpolator PPM Controller Ports
            i_TXPIPPMEN            = 0,
            i_TXPIPPMOVRDEN        = 0,
            i_TXPIPPMPD            = 0,
            i_TXPIPPMSEL           = 1,
            i_TXPIPPMSTEPSIZE      = 0,

            # Transceiver Reset Mode Operation
            i_GTRESETSEL           = 0,
            i_RESETOVRD            = 0,

            # Transmit Ports
            o_TXPMARESETDONE       = Open(),

            # Transmit Ports - Configurable Driver Ports
            i_PMARSVDIN0           = 0b0,
            i_PMARSVDIN1           = 0b0,

            # Transmit Ports - FPGA TX Interface Ports
            i_TXDATA               = self.txdata,
            i_TXUSRCLK             = self.txusrclk,
            i_TXUSRCLK2            = self.txusrclk2,

            # Transmit Ports - PCI Express Ports
            i_TXELECIDLE           = txelecidle,
            i_TXMARGIN             = 0b000,
            i_TXRATE               = 0b000,
            i_TXSWING              = 0,

            # Transmit Ports - Pattern Generator Ports
            i_TXPRBSFORCEERR       = 0,

            # Transmit Ports - TX 8B/10B Encoder Ports
            i_TX8B10BBYPASS        = 0,
            i_TXCHARDISPMODE       = 0,
            i_TXCHARDISPVAL        = 0,
            i_TXCHARISK            = self.txcharisk,

            # Transmit Ports - TX Buffer Bypass Ports
            i_TXDLYBYPASS          = 1 if tx_buffer_enable else 0,
            i_TXDLYEN              = 0 if tx_buffer_enable else tx_init.txdlyen,
            i_TXDLYHOLD            = 0,
            i_TXDLYOVRDEN          = 0,
            i_TXDLYSRESET          = tx_init.txdlysreset,
            o_TXDLYSRESETDONE      = tx_init.txdlysresetdone,
            i_TXDLYUPDOWN          = 0,
            i_TXPHALIGN            = tx_init.txphalign,
            o_TXPHALIGNDONE        = tx_init.txphaligndone,
            i_TXPHALIGNEN          = 0 if tx_buffer_enable else 1,
            i_TXPHDLYPD            = 0,
            i_TXPHDLYRESET         = 0,
            i_TXPHINIT             = tx_init.txphinit,
            o_TXPHINITDONE         = tx_init.txphinitdone,
            i_TXPHOVRDEN           = 0,

            # Transmit Ports - TX Buffer Ports
            o_TXBUFSTATUS          = Open(),

            # Transmit Ports - TX Buffer and Phase Alignment Ports
            i_TXSYNCALLIN          = 0,
            o_TXSYNCDONE           = Open(),
            i_TXSYNCIN             = 0,
            i_TXSYNCMODE           = 0,
            o_TXSYNCOUT            = Open(),

            # Transmit Ports - TX Configurable Driver Ports
            o_GTPTXN               = pads.tx_n,
            o_GTPTXP               = pads.tx_p,
            i_TXBUFDIFFCTRL        = 0b100,
            i_TXDEEMPH             = 0,
            i_TXDIFFCTRL           = 0b1000,
            i_TXDIFFPD             = 0,
            i_TXINHIBIT            = 0,
            i_TXMAINCURSOR         = 0b0000000,
            i_TXPISOPD             = 0,

            # Transmit Ports - TX Fabric Clock Output Control Ports
            o_TXOUTCLK             = self.txoutclk,
            o_TXOUTCLKFABRIC       = Open(),
            o_TXOUTCLKPCS          = Open(),
            i_TXOUTCLKSEL          = 0b010 if tx_buffer_enable else 0b011 if gen == "gen2" else 0b100,
            o_TXRATEDONE           = Open(),

            # Transmit Ports - TX Gearbox Ports
            o_TXGEARBOXREADY       = Open(),
            i_TXHEADER             = 0,
            i_TXSEQUENCE           = 0,
            i_TXSTARTSEQ           = 0,

            # Transmit Ports - TX Initialization and Reset Ports
            i_TXPCSRESET           = 0,
            i_TXPMARESET           = 0,
            o_TXRESETDONE          = tx_init.txresetdone,

            # Transmit Ports - TX OOB signaling Ports
            o_TXCOMFINISH          = txcomfinish,
            i_TXCOMINIT            = txcominit,
            i_TXCOMSAS             = 0,
            i_TXCOMWAKE            = txcomwake,
            i_TXPDELECIDLEMODE     = 0,

            # Transmit Ports - TX Polarity Control Ports
            i_TXPOLARITY           = self.tx_polarity,

            # Transmit Ports - TX Receiver Detection Ports
            i_TXDETECTRX           = 0,

            # Transmit Ports - pattern Generator Ports
            i_TXPRBSSEL            = 0b000,
        )
        self.specials += Instance("GTPE2_CHANNEL", **gtp_params)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from litesata.common import *
from litesata.core import LiteSATACore
from litesata.frontend.arbitration import LiteSATACrossbar
from litesata.frontend.bist import LiteSATABISTGenerator, LiteSATABISTChecker
from litesata.frontend.raid import LiteSATAStriping

from litex.soc.interconnect.stream_sim import *

from test.model.hdd import *

class TestStriping(unittest.TestCase):
    def test_striping(self):
        def generator(dut):
            dut.hdd0.malloc(0, 64)
            dut.hdd1.malloc(0, 64)
            sector = 0
            count = 1
            for i in range(2):
                # Write data
                yield dut.generator.sector.eq(sector)
                yield dut.generator.count.eq(count)
                yield dut.generator.start.eq(1)
                yield
                yield dut.generator.start.eq(0)
                yield
                while not (yield dut.generator.done):
                    yield

                # Verify data
                yield dut.checker.sector.eq(sector)
                yield dut.checker.count.eq(count)
                yield dut.checker.start.eq(1)
                yield
                yield dut.checker.start.eq(0)
                yield
                while not (yield dut.checker.done):
                    yield
                errors = (yield dut.checker.errors)
                print("errors {}".format((yield dut.checker.errors)))
                self.assertEqual(errors, 0)

                # Prepare next iteration
                sector += 1
                count = count + 1

        class DUT(Module):
            def __init__(self):
                self.submodules.hdd0 = HDD(n=0,
                        link_debug         = False,
                        link_random_level  = 0,
                        transport_debug    = False,
                        transport_loopback = False,
                        hdd_debug          = True)
                self.submodules.core0 = LiteSATACore(self.hdd0.phy)

                self.submodules.hdd1 = HDD(n=1,
                        link_debug         = False,
                        link_random_level  = 0,
                        transport_debug    = False,
                        transport_loopback = False,
                        hdd_debug          = True)
                self.submodules.core1 = LiteSATACore(self.hdd1.phy)

                self.submodules.striping = LiteSATAStriping([self.core0, self.core1])
                self.submodules.crossbar = LiteSATACrossbar(self.striping)

                self.submodules.generator = LiteSATABISTGenerator(self.crossbar.get_port())
                self.submodules.checker   = LiteSATABISTChecker(self.crossbar.get_port())

        dut = DUT()
        generators = {
            "sys" :   [generator(dut),
                       dut.hdd0.link.generator(),
                       dut.hdd0.phy.rx.generator(),
                       dut.hdd0.phy.tx.generator(),
                       dut.hdd1.link.generator(),
                       dut.hdd1.phy.rx.generator(),
                       dut.hdd1.phy.tx.generator()]
        }
        clocks = {"sys": 10}
        run_simulation(dut, generators, clocks)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2016 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from litesata.common import *
from litesata.core import LiteSATACore
from litesata.frontend.arbitration import LiteSATACrossbar
from litesata.frontend.bist import LiteSATABISTGenerator, LiteSATABISTChecker

from litex.soc.interconnect.stream_sim import *

from test.model.hdd import *

class TestBIST(unittest.TestCase):
    def test_bist(self):
        def main_generator(dut):
            dut.hdd.malloc(0, 64)
            sector    = 0
            count     = 1
            generator = dut.generator
            checker   = dut.checker
            for i in range(4):
                # Write data
                yield dut.generator.sector.eq(sector)
                yield dut.generator.count.eq(count)
                yield dut.generator.start.eq(1)
                yield
                yield dut.generator.start.eq(0)
                yield
                while not (yield generator.done):
                    yield

                # Verify data
                yield dut.checker.sector.eq(sector)
                yield dut.checker.count.eq(count)
                yield dut.checker.start.eq(1)
                yield
                yield dut.checker.start.eq(0)
                yield
                while not (yield dut.checker.done):
                    yield

                errors = (yield dut.checker.errors)
                print("errors {}".format(errors))
                self.assertEqual(errors, 0)

                # Prepare next iteration
                sector += 1
                count = max((count + 1)%8, 1)

        class DUT(Module):
            def __init__(self, dw=32):
                self.submodules.hdd = HDD(
                    link_debug         = False,
                    link_random_level  = 0,
                    transport_debug    = False,
                    transport_loopback = False,
                    hdd_debug          = True)
                self.submodules.core      = LiteSATACore(self.hdd.phy)
                self.submodules.crossbar  = LiteSATACrossbar(self.core)
                self.submodules.generator = LiteSATABISTGenerator(self.crossbar.get_port(dw))
                self.submodules.checker   = LiteSATABISTChecker(self.crossbar.get_port(dw))

        dut = DUT()
        generators = {
            "sys" : [main_generator(dut),
                       dut.hdd.link.generator(),
                       dut.hdd.phy.rx.generator(),
                       dut.hdd.phy.tx.generator()]
        }
        clocks = {"sys": 10}
        run_simulation(dut, generators, clocks)
